/*******************************************************************************
 serverjs_handlers.js: common socket.io message handlers, as a "require" module
 
 Usage:

 $Author: dmc $
 $Revision: 3206 $
 $Date: 2023-07-06 10:53:57 -0300 (Thu, 06 Jul 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/serverjs_handlers.js $
*******************************************************************************/

var log4js = require('log4js');
var _log = log4js.getLogger('serverjs_handlers.js');

var fs = require("fs"); // file system
var _has = require('lodash.has');
//// for log():
//eval(fs.readFileSync("./smt.common.lib/public/javascripts/utils.js")+""); // common utils

var new_storagejs = require('./new_storage.js');
var _performance = require( "./lib/utils/performance.js" );
var mil_serverjs_handlers;
try {
	mil_serverjs_handlers = require('./mil_serverjs_handlers.js');
} catch(e) {
}

var http = require("http");

module.exports.export_scores_csv_string = function (req) {
// loop through archive and save all strings not yet saved in database
//function rstat_save_scores(socket, next) {
	
	var match_id = req.params.match_id;
	var count = 0;
	var out_str = '';
	out_str += 'String_ID, EVNTNO, EVTName, MtchNo, MatchName, StgNo, StageName, COMPID, COMPName, NShots, ScoreDecimal, ScoreText\n';
	
	function sanitize(str) { // make str safe to use as a CSV field
		if ('undefined' == typeof str) return "";
		if ('number' == typeof str) return str;
		if ('string' != typeof str) {
			str = JSON.stringify(str);
		}
		return str.replace(/,/g,';');
	}
	
	for (var i in archive) {
		var string = archive[i]; // no shots

		// skip if desired
		if (string.rstat_string_action == "abandon") continue;

		if (match_id) { // restrict to only match_id
			if (!string.event_data) {
				continue;
			}
			if (string.event_data && string.event_data.event_id != match_id) {
				continue;
			}
		}


		// resolve note and continue
		if (string.rstat_string_action == "resolve") {
			string.note_resolved = true;
		}

		// shooter count is always set on saved strings
		for (var i = 0; i < string.shooter_count; i++) {

			// requires valid user assigned
			var user_id = 0;
			if (string.role_shooter && string.role_shooter[i]) {
				user_id = string.role_shooter[i];
			} else {
//				continue;
			}

			// must exist and be from rstat database
			var user_id_str = "" + user_id;
//			if (!data.users[user_id] || user_id_str.substring(0, 13) != "RSTAT_COMPNO_") continue;

			// must be final
//			if (string.score[i].count != string.shots_mode) continue;
			
//			log ("saving score " + string.score[i].string + " for user " + data.users[user_id].realname);
			
			count++;
			var user_id = string.role_shooter ? string.role_shooter[i] : 0;
			var score = string.score ? string.score[i] : undefined;
			

			out_str += 
				string._id
				+ "," + (string.event_data ?
//				string.event_data.event_id + ',' + data.events[string.event_data.event_id].name.replace(/,/g,';')
				string.event_data.event_id + ',' + sanitize(data.events[string.event_data.event_id].name)
				+ "," + string.event_data.match_id
//				  + ',' + data.events[string.event_data.event_id].matches[string.event_data.match_id].name.replace(/,/g,';')
				  + ',' + sanitize(data.events[string.event_data.event_id].matches[string.event_data.match_id].name)
				+ "," + string.event_data.stage_id
//				  + ',' + data.events[string.event_data.event_id].matches[string.event_data.match_id].stages[string.event_data.stage_id].name.replace(/,/g,';')
				  + ',' + sanitize(data.events[string.event_data.event_id].matches[string.event_data.match_id].stages[string.event_data.stage_id].name)
				: '-,-,-,-,-,-')
//				 + "," + (user_id ? data.users[user_id].name.replace(/,/g,';') + ',' + (data.users[user_id].realname).replace(/,/g,';') : "noname,-")
				 + "," + (user_id ? sanitize(data.users[user_id].name) + ',' + sanitize((data.users[user_id].realname)) : "noname,-")
				 + "," + (score ? score.count + ',' + score.rstat_score + ',' + score.string : "-,noscore,-")
				 + "\n";

		}
		
	}

	return out_str;

}


module.exports.attach_app_handlers = function (app) {

	app.get("/archive",get_archive);

	app.get("/download/strings-csv/:match_id?", function (req, res, next) {

		_log.debug("GET /download/strings-csv");
/*		var ids = req.params.ids.split( "," );
*/
//		res.attachment('strings.csv');
//res.set('Content-Type', 'application/octet-stream');
//		res.contentType('csv');

		res.end(module.exports.export_scores_csv_string(req));
	} );

	app.get("/download/users-csv", function (req, res, next) {

//		res.contentType('csv');
//res.set('Content-Type', 'application/octet-stream');

		res.write('a,b,c');
		res.write('1,2,3');
		res.end();

	} );

	// restore all data from a zip
	app.post("/import-csv", function (req, res, next) {

		var file = req.files["selected_file"];
		_log.debug("/import-csv",req.files);
		if (file) { // accept all file types, we'll figure out what to do...
			// https://stackoverflow.com/questions/6156501/read-a-file-one-line-at-a-time-in-node-js/15554600#15554600
//			const fs = require('fs');
//			const src = fs.createReadStream(file.path);
//			const readline = require('readline');
//			const lineReader=readline.createInterface(file.ws);
//			const lineReader=readline.createInterface({
//				input: fs.createReadStream(file.path)
//			});
//console.log("SMC lineReader=",lineReader);
//			res.writeHead(200, { 'Content-Type': 'text/html' });

			_log.debug('test fast-csv');
			var csv = require("fast-csv");
			csv
				.fromPath(file.path, {headers : true})
				.on("data", function(data){
					_log.debug("data--",data);
 				})
 				.on("end", function(){
 					_log.debug("done");
 				});

//console.log("DMC src=",src);
//src.pipe(res);
//			file.ws.write(res);
			res.end();
		} else {
			res.writeHead(404, { 'Content-Type': 'text/html' });
			res.write("<h1>File not found, or not of type '.tgz'");
			res.write("<p><button class='title_element' onclick='window.history.back()'>BACK");
			res.end();
		}
	});



	app.get("/download/reports-csv/:ids", function (req, res, next) { // SOLO
		var ids = req.params.ids.split( "," );
		new_storagejs.storage_write_report(ids);
		res.write( "reports/report.csv" );
		res.end();
	} );

// http://localhost:4080/download/string-json/1523816698519
	app.get("/download/string-json/:id", function (req, res, next) { // S25
		var id = req.params.id;
		var start = _performance();
		new_storagejs.storage_read_string_async(id,function(err,str){
			if (!err) {
				res.json(str || {});
			} else { // error of some sort
				res.json({}); // error
			}
		});
	} );

	// restore all data; "_new" is for FA
	app.post("/restore_all_data_new", function (req, res, next) {
		var file = req.files["selected_file"];
		if (file) { // accept all file types, we'll figure out what to do...
			new_storagejs.storage_restore_all_data( file.path, function( err, stdout, stderr ) {
				var msg="";
				if (err) {
					msg += "<h1>ERROR!!</h1><p><h2><pre>"+err.toString() +"</pre></h2>";
				} else {
					msg += "<h1> SUCCESS</h1>";
					msg += "<p><h2>Restarting & Reloading...</h2>";
				}
				if (stdout) msg += "<p><h3>stdout:<p><pre style='margin-left:4em;'>"+stdout+"</pre></h3>";
				if (stderr) msg += "<p><h3>stderr:<p><pre style='margin-left:4em;'>"+stderr+"</pre></h3>";
				res.json( {
					valid: !err,
					msg: msg
				} );
				if ( !err ) { // we need to restart ourselves..
					process.nextTick( function() {
						process.exit(0); // should restart itself
					} );
				}
			} );
		} else {
			res.json( { 
				valid: false,
				msg: "File not found, or not of type '.tgz'"
			} );
		}
	} );




	// restore all data from a zip
	app.post("/restore_all_data", function (req, res, next) {
	//	res.redirect("back");

		var file = req.files["selected_file"];
//		if (file && file.originalFilename.indexOf(".tgz") >= 0) {
		if (file) { // accept all file types, we'll figure out what to do...
			res.writeHead(200, { 'Content-Type': 'text/html' });
			_log.debug("/restore_all_data",file.path,file.originalFilename);
			new_storagejs.storage_restore_all_data(file.path,function(err, stdout, stderr) {
				if (err) {
					res.write("<h1>ERROR!!</h1><p><h2><pre>"+err.toString() +"</pre></h2>");
				} else {
					res.write("<h1> SUCCESS</h1>");
					res.write("<p><h2>Restarting & Reloading...</h2>");
				}
				if (stdout) res.write("<p><h3>stdout:<p><pre style='margin-left:4em;'>"+stdout+"</pre></h3>");
				if (stderr) res.write("<p><h3>stderr:<p><pre style='margin-left:4em;'>"+stderr+"</pre></h3>");
				res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
				if (!err) { // we need to restart ourselves..
					process.nextTick(function() {
						process.exit(0); // should restart itself
					});
				}
				res.end();
			});
		} else {
			res.writeHead(404, { 'Content-Type': 'text/html' });
			res.write("<h1>File not found, or not of type '.tgz'");
			res.write("<p><button class='title_element' onclick='window.history.back()'>BACK");
			res.end();
		}
	});

	app.get('/firmware-update-warn', function (req, res) { // courtesy warning that we are about to have a firmware file uploaded to us
		_log.debug('/firmware-update-warn start');
		var start = _performance();
		var have_gc = false;
		var mem = process.memoryUsage();
		var mem_before = "<p>before: rss="+mem.rss+", heapTotal="+mem.heapTotal+", heapUsed="+mem.heapUsed+"\n";

		if (global.gc) {
			_log.debug("starting gc()");
			have_gc = true;
			global.gc();
			_log.debug("finished gc(), now ");
		} else {
			_log.debug("global.gc not defined");
		}
		var mem = process.memoryUsage();
		var mem_after = "<p>after: rss="+mem.rss+", heapTotal="+mem.heapTotal+", heapUsed="+mem.heapUsed+"\n";
		res.writeHead(200, { 'Content-Type': 'text/html' });
		res.write(
			"<h1>Done</h1>\n"
			+"<p>have_gc="+have_gc+"\n"
			+mem_before+mem_after
			+"<p>time="+ _performance(start)+"ms"
		);
		res.end();

	});
	// validate the uploaded firmware file
	app.post('/firmware-update-validate', function (req, res) { 
		var file = req.files['selected_file'];
		var valid = false;
		var msg = '';
		_log.debug('/firmware-update-validate',file&&file.ws ? file&&file.ws.bytesWritten : "unknown","bytes");	
		if (file && file.originalFilename !== '') {
			valid = true;
			var exec = require("child_process").exec;
			var cmd = 'sysupgrade -T ' + file.path;
			var mem = process.memoryUsage();
			var mem_before = "<p>before: rss="+mem.rss+", heapTotal="+mem.heapTotal+", heapUsed="+mem.heapUsed+"\n";
			if (global.gc) {
				_log.debug("starting gc()",mem_before);
				global.gc();
				var mem = process.memoryUsage();
				var mem_after = "<p>after: rss="+mem.rss+", heapTotal="+mem.heapTotal+", heapUsed="+mem.heapUsed+"\n";
				_log.debug("finished gc()",mem_after);
			} else {
				_log.debug("global.gc not defined",mem_before);
			}
			exec(cmd, function(error, stdout, stderr) {
				_log.debug('/firmware-update-validate sysupgrade -T =',error,stdout);
				if (error) {
					msg = '';
					//if (stdout) msg += stdout + '\n';
					if (stderr) msg += stderr + '\n';
					valid = false;
					exec("rm -f " + file.path); // delete uploaded firmware file...
				}
				res.json({ valid: valid, file: file.path, original: file.originalFilename, msg: msg });
			});
		} else {
			res.json({ valid: valid, file: file.path, original: file.originalFilename, msg: msg });
		}
	});

	// executes the firmware update
	app.get('/firmware-update-execute/:filename', function(req, res) {
		var filename = req.params.filename;
	
		var exec = require("child_process").exec;
		var cmd = 'sysupgrade ' + filename;
		_log.debug('run command: ' + cmd);
		exec(cmd, function(error, stdout, stderr) {
			// console.log(error);
			// console.log(stdout);
			// console.log(stderr);
		});

		res.json({ file: filename });
	});

	app.get('/firmware-update-cancel', function(req, res) {
		var exec = require("child_process").exec;
		exec("rm -f /tmp/*.bin"); // klugey, and far from 100%
		res.json({ success: ok });
	});

	// data export function that works with circulare data (which can't use JSON.stringify)
	app.get('/global/:member?/:depth?', function(req, res) { // e.g.  /global/fa_clients/2
		var depth = req.params.depth;
		var member = req.params.member || 'global';
		var start = _performance();
		var util = require('util')
		var tgt = (global && global[member]) || {};
		var str = util.inspect(tgt, {showHidden: false, depth: depth}); // circular structure, can't use JSON.stringify
		res.end(str); // circular structure, can't use JSON.stringify
		_log.debug('/global - done (' + _performance(start) + 'ms)', str.length,'bytes');
	});



	app.get("/download-csv/users", function (req, res, next) { // SOLO
//		res.attachment('users.csv');
		var fast_csv=require('fast-csv');
		var csvStream = fast_csv.format({
		//headers:true
			headers : ["_id","name","realname","bullet_dia", "fclass_target","password","is_range_officer"]
		});
		csvStream.pipe(res);
		for (var i in data.users) {
			csvStream.write(data.users[i]);
		}
		csvStream.end();

	} );

	app.post('/test-echo-post', function(req, res) {
		res.json(req.body);
		res.end();
	});

	app.get('/test-echo-get', function(req, res) {
		var response = {
			'message-received': 'GET /test-echo-get',
			'timestamp' : Date()
		};
		res.json(response);
		res.end();
	});

	app.post('/replay_shot', function(req, res) {
		fn_replay_shot(req.body.shotdata, req.body.frame_name);
		res.json({}); // TODO: some sort of status?
		res.end();
	});
	app.post('/import-json/:membername', function(req, res) {
	
		var start = _performance();
		var membername = req.params.membername;
		var crud_msg = { };
		var crud_msg_membername = crud_msg[membername] = {}
		// id: events
		// _id: faces, frames, strings, users
		// internalId: devices
		var id_name_table = { // field name for data types that don't use default "_id"
			"events": "id",
			"devices": "internalId",
		};
		var id_name = id_name_table[membername] || "_id"; // default
		// accept array or object in req.body
		if ('object' == typeof req.body) {
			for (var k in req.body) {
				var datum = req.body[k];
				// to DELETE an object: "1552987027284" : null
_log.debug("/import-json",'key','"'+k+'"',typeof k,id_name);
				if (datum == null) { // 
					datum = g_special_delete_obj; // { _delete: true };
					crud_msg_membername[k] = datum;
					_log.debug("key",k,"null ---> delete");
				} else {
					if (!datum[id_name] || datum[id_name] == -1) { // create an id
						datum[id_name] = new_storagejs.storage_get_unique_id();
					}
					crud_msg_membername[datum[id_name]] = datum;
				}
			}
		}
_log.debug("/import-json crud_msg=",crud_msg);
		res.json(crud_msg);
		_log.debug('/import-json - processing done (' + _performance(start) + 'ms)');
		apply_crud(crud_msg);
		_log.debug('/import-json - crud done (' + _performance(start) + 'ms)');
	});



	app.get('/data/:membername', function(req, res) { // frames, strings, devices, config, users, events, faces
		var membername = req.params.membername;
		_log.debug("GET /data/:membername",membername);
		res.json(data[membername] || {});
	});

	// send all data differently...
	app.get("/data", function(req, res) {
		_log.debug("GET /data - (seq " + sequence_id + ")");
		var start = _performance();
		data.seq = sequence_id;
		data.ts = Date.now();
		data.interval = heartbeat_n_fast_deltas * fast_delta_interval; // normally 1500ms

		// try some caching
		var data_string = "";
		if (!cached_data_string || cached_data_id != sequence_id) {
		    _log.debug('data_string cache miss, stringify current data object (sequence_id: ' + sequence_id + ')');
			data_string = JSON.stringify(data); // on S50 this takes about 90ms for 389KB
			cached_data_string = data_string;
			cached_data_id = sequence_id;
		    _log.debug('data_string finished stringfy ('+_performance(start)+ 'ms, sequence_id: ',sequence_id, 'bytes:',data_string.length,')');
		} else {
		    _log.debug('data_string cache hit (' + _performance(start) + 'ms, sequence_id: ',sequence_id,')');
			data_string = cached_data_string;
		}
		// use express.end() instead of express.send()
		// @see: https://stackoverflow.com/a/15564379
		// @see: https://stackoverflow.com/a/22774872
		//res.send(data_string); // i7: 13..33; ON S50: 102..215 for 388KB.  typ. 100ms (which include the stringify's 90ms)
		res.end(data_string);
		_log.debug('finished response data (' + _performance(start) + 'ms)');
	});


	// GET /set_altnetwork on an S7: do GET 192.168.0.7:/set_altnetwork and local (uci set wireless...)
	// GET /set_altnetwork on an S25: same
	app.get("/set_altnetwork/:network?", function (req, res) {
		// 2021-Mar-16: SSID replacement-suffix may be "", or any other value (formerly 1..4 only)
		var network = req.params.network || ""; // undefined --> ""
		var exec = require("child_process").exec;

		exec( "sh ./set_altnetwork.sh " + network, function(err,stdout,stderr) {
			if (err) {
				console.log("err",err);
				console.log("err.message",err.message);
				res.end(err.message);
			} else {
				res.end(stdout);
			}
		});
	});

/*

root@SMT_S50_server:~# cat /tmp/dhcp.leases
1618447171 c4:93:00:0b:52:f2 192.168.0.249 * *
1618448711 c4:93:00:0b:53:21 192.168.0.239 dwq *
1618430836 e4:8d:8c:f4:a0:b1 192.168.0.197 MikroTik 01:e4:8d:8c:f4:a0:b1
1618436088 a0:88:69:9d:c3:8f 192.168.0.219 SMT-LT-14 01:a0:88:69:9d:c3:8f


*/

	app.get("/tsolo_recover", function (req, res) {
		// 2021-Mar-16: SSID replacement-suffix may be "", or any other value (formerly 1..4 only)
		var network = req.params.network || ""; // undefined --> ""
		var exec = require("child_process").exec;

		exec("./tsolo_recover.sh " + network, function(err,stdout,stderr) {

			if (err) {
				res.write("<h1>ERROR!!</h1><p><h2><pre>"+err.toString() +"</pre></h2>");
			} else {
				res.write("<h1> SUCCESS</h1>");
			}
			if (stdout) res.write("<p><h3>stdout:<p><pre style='margin-left:4em;'>"+stdout+"</pre></h3>");
			if (stderr) res.write("<p><h3>stderr:<p><pre style='margin-left:4em;'>"+stderr+"</pre></h3>");
			res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
			res.end();
		});
	});

	app.get("/s75g", function (req, res) {
		var exec = require("child_process").exec;

		exec( "./s7_fixup_5g.sh ", function(err,stdout,stderr) {
			if (err) {
				res.write("<h1>ERROR!!</h1><p><h2><pre>"+err.toString() +"</pre></h2>");
			} else {
				res.write("<h1> SUCCESS</h1>");
			}
			if (stdout) res.write("<p><h3>stdout:<p><pre style='margin-left:4em;'>"+stdout+"</pre></h3>");
			if (stderr) res.write("<p><h3>stderr:<p><pre style='margin-left:4em;'>"+stderr+"</pre></h3>");
			res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
			res.end();
		});
	});
	app.get("/set_wifi_password/:password?", function (req, res) {
 		var exec = require("child_process").exec;
		var password = req.params.password || ""; // undefined --> ""

		exec( "./set_wifi_password.sh '"+password+"'", function(err,stdout,stderr) {
			if (err) {
				res.write("<h1>ERROR!!</h1><p><h2><pre>"+err.toString() +"</pre></h2>");
			} else {
				res.write("<h1> SUCCESS</h1>");
			}
			if (stdout) res.write("<p><h3>stdout:<p><pre style='margin-left:4em;'>"+stdout+"</pre></h3>");
			if (stderr) res.write("<p><h3>stderr:<p><pre style='margin-left:4em;'>"+stderr+"</pre></h3>");
			res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
			res.end();
		});
	});

	app.get('/archive-list', function (req, res) {
		_log.debug('called get archive action');
		var result = archive_collection.getListItems(req.query.page, req.query.cacheKey, req.query.numberOfItems);
		var startResponse = _performance();
		res.json(result);
		_log.debug('finished response archive data (' + _performance(startResponse) + 'ms)');
	});


	var last_delta_stringified = undefined;
	var stringified_delta;

	app.get("/delta/:deltano", function(req, res) {
		var start = _performance();
		var deltano = req.params.deltano;
		_log.debug("GET /delta/",deltano);
		var the_delta = deltasSent[deltano];
		if (!the_delta) { //send
			the_delta="";
		}
		res.write(the_delta/*string*/);
		res.end();
		_log.debug('get /delta/'+deltano+' (' + _performance(start) + 'ms) len=',the_delta.length);
	});

	app.get('/data-readable', function(req, res) { // slow & pretty, for testing/dev use only
		var start = _performance();
		res.json(data || {});
		_log.debug('/data-readable - res.json done (' + _performance(start) + 'ms)');
	});


	// import targetfaces
	app.post("/import_targetfaces_new", fn_import_targetfaces_new);

	app.post("/import_targetfaces", fn_import_targetfaces);

	app.post("/calc_shot_timing", function (req, res, next) {

	//console.log(Object.keys(req));
	// .. snippet of interesting-looking fields:
	//  'socket',
	//console.log("DMC /calc_shot_timing 1 req.socket")
	//console.log(req.socket);
	//  'connection',
	//console.log("DMC /calc_shot_timing 1 req.connection")
	//console.log(req.connection);
	//  'httpVersion',
	//  'complete',
	//console.log("DMC /calc_shot_timing req.complete")
	//console.log(req.complete); // boolean
	//  'headers',
	//console.log("DMC /calc_shot_timing req.headers")
	//console.log(req.headers);
	//  'trailers',
	//  'url',
	//  'method',
	//  'statusCode',
	//  'client',
	//  'originalUrl',
	//  'query',
	//console.log("DMC /calc_shot_timing 1 req.query")
	//console.log(req.query);
	//  'res',
	//  'next',
	//  'body',
	//console.log("DMC /calc_shot_timing 1 req.body")
	//console.log(req.body);
	//  'files',
	//console.log("DMC /calc_shot_timing req.files")
	//console.log(req.files); // can be undefined
	//  'route',
	//console.log("DMC /calc_shot_timing 1 req.route")
	//console.log(req.route);
	/*
		DMC /calc_shot_timing 1 req.route
		{ path: '/calc_shot_timing',
		  method: 'post',
		  callbacks: [ [Function] ],
		  keys: [],
		  regexp: /^\/calc_shot_timing\/?$/i,
		  params: [] }
		*/
	//  'params' ]
	//console.log("DMC /calc_shot_timing 1 req.params")
	//console.log(req.params);
	// ---> []

		var calcd_timings = namedframe_calc_timings_sim_shot(req); // req.body.{frame_name, x, y, v, theta, phi, temp} needs to be supplied

	  res.send(calcd_timings || {});    // send result back; could be empty if parm errors.
	  res.end();
	});




	//app.post("/namedframe_emit_shot", namedframe_emit_shot_req_res);
	app.post("/namedframe_emit_shot", function(req,res,net) {
		namedframe_emit_shot_req(req);
		res.end();
	});

	app.get("/smt_on", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_SMT: true};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Show extended SMT diagnostics/features");
	});

	app.get("/smt_off", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_SMT: false};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Hide extended SMT diagnostics/features");
	});

	app.get("/nra_on", function(req, res) { // DMC 4-July-2022
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_NRA: true};
		apply_crud(crud); // apply crud message
		report_config_change(res, "NRA-2022 Atterbury mode ON");
	});

	app.get("/nra_off", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_NRA: false};
		apply_crud(crud); // apply crud message
		report_config_change(res, "NRA-2022 Atterbury mode OFF");
	});

	app.get("/toast_stars/:onoff?", function(req, res) { // DMC 9-July-2022
		var onoff = req.params.onoff || ""; // undefined --> ""
		if (onoff == "") onoff = data.config.toast_stars ? "off" : "on"; // toggle
		if (onoff != "off") onoff = "on"
		var crud = {}; // build crud message to apply change to active data
		crud.config = {toast_stars: onoff == "on"};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Popup messages for starred shots ---> " + onoff);
	});

	app.get("/rstat_on", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_rstat: true};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Show RSTAT sync button");
	});

	app.get("/rstat_off", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_rstat: false};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Hide RSTAT sync button");
	});

	app.get("/health_on", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_health: true};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Show wifi health diagnostics");
	});

	app.get("/health_off", function(req, res) {
		var crud = {}; // build crud message to apply change to active data
		crud.config = {is_health: false};
		apply_crud(crud); // apply crud message
		report_config_change(res, "Hide wifi health diagnostics");
	});
	
	function report_config_change(res, msg) {
		res.writeHead(200, { 'Content-Type': 'text/html' });
		res.write("<h1>Configuration change</h1><p>Now: "+msg);
		res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
		res.end();
	}

	mil_serverjs_handlers && mil_serverjs_handlers.attach_app_handlers(app);

	app.get("/favicon.ico", function(req, res) { // never called;
		_log.debug("app.get /favicon.ico handler", req.url);
	});

	app.get("*", function(req, res, next) {
		_log.debug("app.get * handler", req.url);
		res.writeHead(404, { 'Content-Type': 'text/html' });
		res.write("<h1>Unrecognized URL:</h1>"+req.url);
		res.write("<p><button class='title_element' onclick='window.history.back()'>BACK</button>");
		res.end();
	});

}

//attach_socketio_handlers
module.exports.attach_socketio_handlers = function (socket,io) {

	mil_serverjs_handlers && mil_serverjs_handlers.attach_socketio_handlers(socket,io);

	// load an inactive string and send it
	socket.on("get_string", function(msg) {
	var start = _performance();
		var str = new_storagejs.storage_read_string_async(msg.string_id,function(err,str){
			if (!err && str) {
				// log("sending old string to client", msg.string_id);
				//console.log('get_string() - callback started (' + _performance(start) + 'ms)','str.length=',str.length);
				socket.emit("string", str); // TODO: migrate to GET /....
				//console.log('get_string() - emit done (' + _performance(start) + 'ms)');
			}
		});
	});

	socket.on("replay_string",fn_replay_string); // replay a whole string
	socket.on("replay_shot",fn_replay_shot); // replay a single shot
	socket.on("insert_missed_shot", fn_insert_missed_shot);
	// not quite ready yet to merge create_string, but that will be good!
//	socket.on("create_string",create_string);



	socket.on("client_refresh", function() {
		_log.debug("all clients refreshing...");
		io.emit("client_refresh");
	});
	
	socket.on("download_shotslog", function(days) {
		if ('object' == typeof global.archive_collection) { // FA, G2
			new_storagejs.storage_write_shotslog(days, archive_collection.getItems());// FA, G2
		} else {
			new_storagejs.storage_write_shotslog(days, archive); // SMTV2-LODI - sets global.archive
		}
		socket.emit("download_file", "shotslog.csv");
	});

	socket.on("download_report", function(report_strings) { // SMTV2-LODI, G2
		new_storagejs.storage_write_report(report_strings);
		socket.emit("download_file", "reports/report.csv");
	});


	socket.on("download_shot", function(shot) { // TODO still to be implemented
//		storage_write_report(report_strings);
		socket.emit("download_file", "shot.csv");
	});

	socket.on("crud", fn_socket_on_crud);
	socket.on("SOLO_C2_sw_update", get_fn_socket_on_SOLO_C2_sw_update(socket));
	socket.on("SOLO_C2_sw_update2", get_fn_socket_on_SOLO_C2_sw_update2(socket));
	socket.on("SOLO_C2_reset_medium", get_fn_SOLO_C2_reset_medium(socket));
	socket.on("SOLO_C2_reset_light", fn_SOLO_C2_reset_light);
	socket.on("SOLO_C2_reset2_medium", get_fn_SOLO_C2_reset2_medium(socket)); // via :280
	socket.on("SOLO_C2_reset2_heavy", get_fn_SOLO_C2_reset2_heavy(socket)); // via :280
}



var propertiesApplyCrud = [
	'frames.SOLO.geom.mic_width',
	'frames.SOLO.geom.x_cal',
	'faces',
	'config' // DMC: this hits nearly EVERYTHING on GEARWHEEL page; need to improve this!!!
];

function is_empty(obj) { // https://stackoverflow.com/questions/679915/how-do-i-test-for-an-empty-javascript-object
	if ('object' != typeof obj) return false;
	for(var prop in obj) {
		if(obj.hasOwnProperty(prop)) {
			return false;
		}
	}

	return JSON.stringify(obj) === JSON.stringify({});
}

// tar cf - TSOLO-Carambola | time ssh  -y -i ~/.ssh/id_rsa2 root@192.168.0.217 'cd /www/smtv2; pwd; id; tar xf - '
//   64 ssh -y root@::ffff:192.168.0.47

function get_fn_socket_on_SOLO_C2_sw_update(socket,msg_parms) {
	return function (msg_parms) {
		var device_id = msg_parms.device_id;
		var device_data = global.data.devices[device_id];

		var cmd = 'cd /www/smtv2/TSOLO-Carambola && tar cf - . | ';
		cmd += 'ssh  -y -i /root/.ssh/id_rsa2 root@' + device_data.ping_health.my_ip;
		cmd += ' " cd /www/smtv2 && tar xf - && (cp cgi-bin/* /www/cgi-bin; true) && /etc/init.d/smt-server restart && ls -l"';
		
		cmd = "./SOLO_C2_sw_update.sh " + device_data.ping_health.my_ip;
		exec_and_log(socket,device_id,cmd,"SOLO_C2_sw_update");
	}
}

function get_fn_socket_on_SOLO_C2_sw_update2(socket,msg_parms) {
	return function (msg_parms) {
		var device_id = msg_parms.device_id;
		var device_data = global.data.devices[device_id];

		var cmd = 'cd /www/smtv2/TSOLO-Carambola && tar cf - . | ';
		cmd += 'ssh  -y -i /root/.ssh/id_rsa2 root@' + device_data.ping_health.my_ip;
		cmd += ' " cd /www/smtv2 && tar xf - && (cp cgi-bin/* /www/cgi-bin; true) &&';
		cmd += ' " mv server2.js server.js &&'; // have "server2.js" override
		cmd += ' /etc/init.d/smt-server restart && ls -l"';
		exec_and_log(socket,device_id,cmd,"SOLO_C2_sw_update2");
	}
}




function get_IP_C2_parms(msg_parms) { // return device IP if available, otherwise false
	var device_id = msg_parms.device_id;
	if (!device_id) return false;
	var device_data = global.data.devices[device_id];
	if (!device_data || !device_data.ping_health || !device_data.ping_health.my_ip) return false;
	// OK:
	return device_data.ping_health.my_ip;
}

function http_280_get(host, cmd, cb) {
	var http_opts = { host: host, port: 280, path: '/cgi-bin/' + cmd }; // defaults: HTTP:, GET
	var req = http.request(http_opts, function(res) {
		var body = "";
		res.on("data", function (chunk) { body += chunk; });
		res.on("end", function () {
			// http_280_get body= <h1>Not Found</h1>The requested URL /cgi-bin/logread was not found on this server.
			var notfound_msg = '<h1>Not Found</h1>';
			if (-1 != body.indexOf(notfound_msg)) {
				cb(notfound_msg,body);
				return;
			}
			cb(null,body); // OK; return html text
		});
		res.on('error', cb );
	});
	req.on('error', cb );
	req.end();
}

function do_and_log_280(my_socket, parmname/*SOLO_C2_reset2_medium*/, commandname/*restart-smt-server*/, msg_parms) {
	//console.log('do_and_log_280',parmname,msg_parms);
	var my_ip = get_IP_C2_parms(msg_parms);
	if (false == my_ip) return;
	var device_id = msg_parms.device_id;

	http_280_get(my_ip,commandname, function(err,result) {
		if (!result) result = "";
		var back_button = "<p><button class='title_element' onclick='window.history.back()'>BACK</button>";
		result=back_button+"<h1>" + device_id + " " + commandname + " " + my_ip + "</h1>" + result+back_button;

		if (err) {
			result += "<h1>Error encountered</h1><PRE>" + err + "</PRE>" + result;
		} else {
			// only on success - reset to indicate "done"...
			var crud = { devices: {} }; // build crud message
			crud.devices[device_id] = {}; // e.g. 3346
			crud.devices[device_id][parmname] = false; // e.g. SOLO_C2_reset2_medium
			fn_socket_on_crud(crud);
		}
		new_storagejs.storage_write_logfile(result,function(error, filepath_saved) {
			// check error??
			console.log(parmname,device_id,"filepath_saved=",filepath_saved/*,"result=",result*/);
			my_socket.emit("download_file", filepath_saved);
		});
	});
}


//SOLO_C2_reset2_medium
function get_fn_SOLO_C2_reset2_medium(socket,msg_parms) {
	return function (msg_parms) {
		do_and_log_280(socket, "SOLO_C2_reset2_medium","restart-smt-server", msg_parms);
	}
}

//SOLO_C2_reset2_heavy
function get_fn_SOLO_C2_reset2_heavy(socket,msg_parms) {
	return function (msg_parms) {
		do_and_log_280(socket, "SOLO_C2_reset2_heavy","reboot-carambola2", msg_parms);
	}
}

function exec_and_log(my_socket,device_id/*3346*/,cmd/*ssh...*/,parmname/*SOLO_C2_reset_medium*/) {
	var exec = require("child_process").exec;
	exec(cmd, function(err, stdout, stderr) {
		console.log(parmname,"exec callback");
		console.log("err=",err);
		var result="<h1>" + parmname + " " + device_id + "</h1><h2>command</h2><PRE>" + cmd + "</PRE>";
		
		if (err) {
			result += "<h1>err=</h1><PRE>" + err + "</PRE>";
		} else {
			// only on success - reset to indicate "done"...
			var crud = { devices: {} }; // build crud message
			crud.devices[device_id] = {};
			crud.devices[device_id][parmname] = false;
			fn_socket_on_crud(crud);
		}
		if (stderr) result += "<h1>STDERR</h1><PRE>" + stderr + "</PRE>";
		if (stdout) result += "<h1>STDOUT</h1><PRE>" + stdout + "</PRE>";

		new_storagejs.storage_write_logfile(result,function(error, filepath_saved) {
			// check error??
console.log(parmname,"filepath_saved=",filepath_saved/*,"result=",result*/);
			my_socket.emit("download_file", filepath_saved);
		});
	});
}


function get_fn_SOLO_C2_reset_medium(socket,msg_parms) {
	return function (msg_parms) {
		console.log("SOLO_C2_reset_medium",msg_parms);
		var device_id = msg_parms.device_id;
		if (!device_id) return;
		var device_data = global.data.devices[device_id];
		if (!device_data) return;
		if (!device_data.ping_health) return;
		if (!device_data.ping_health.my_ip) return;

		var cmd = '';
		cmd += 'ssh  -y -i /root/.ssh/id_rsa2 root@' + device_data.ping_health.my_ip;
		cmd += ' "/etc/init.d/smt-server restart && ls -l"';
		exec_and_log(socket,device_id,cmd,"SOLO_C2_reset_medium");
		return;
	}
}

function fn_SOLO_C2_reset_light(msg_parms) {
console.log("SOLO_C2_reset_light",msg_parms);
	var device_id = msg_parms.device_id;
	if (!device_id) return;
	var device_data = global.data.devices[device_id];
	if (!device_data) return;
	if (!device_data.ping_health) return;
	if (!device_data.ping_health.my_ip) return;
	if ('number' != typeof device_data.idx) return;
	var client = global.fa_clients[device_data.idx];
	if (!client) return;
	if (!client.active || !client.socket) return;
	client.socket.destroy();
	// success or not, reset to indicate "done"...
	var crud = { devices: {} }; // build crud message
	crud.devices[device_id] = {SOLO_C2_reset_light: false};
	fn_socket_on_crud(crud);
}


// handle general crud message; run storage_write_data() or storage_write_faces() if needed
function fn_socket_on_crud(msg) {
	// do some sanity-checking on msg.strings, to avoid "1 string(s) do not have a valid face!" etc
	var g_strings = global.data.strings;
	var strings_to_process = {};
	for (var i in msg.strings) {
		// if we don't recognize the string and it doesn't have a frame or face, ignore it
		if (!g_strings[i] && (!msg.strings[i].frame_id || !msg.strings[i].face_id)) {
			continue;
		}
		if (undefined == i) {
			continue; // ignore key=='undefined'
		}
		strings_to_process[i]=msg.strings[i];
	}
	var orig_strings = msg.strings;
	msg.strings = strings_to_process;
	apply_crud(msg);
	msg.strings = orig_strings; // restore, so we don't surprise the caller!

	var shouldSave = false;
	for ( var i = 0; i < propertiesApplyCrud.length; i++ ) {
		if ( _has( msg, propertiesApplyCrud[ i ] ) ) {
			// DMC 21-June-2019 - must not falsely trigger on empty object "config: {}"
			if (false == is_empty(msg[propertiesApplyCrud[ i ]])) {
				shouldSave = true;
			}
		}
	}
	if ( shouldSave ) {
		storage_write_data( data );
	}
	// special case, if changes made to faces (stored in another file...)
	if ( msg.faces && false == is_empty(msg.faces)) {
		storage_write_faces( data );
	}
}


function save_string(string) { // not yet being used
	
	var frame = data.frames[string.frame_id];

	// copy geom from frame to string
	string.geom = frame.geom;

	// count valid/invalid shots
	var n_valid = 0;
	var n_invalid = 0;
	var shots_to_save = [];
	for (i in string.shots) {
		var shot = string.shots[i];
		if (!shot) continue; // do not count deleted shots
		shots_to_save.push(shot);
		if (shot.valid) n_valid++; else n_invalid++;
	}

	string.shots = shots_to_save;
	string.n_shots = n_valid;
	string.n_invalid = n_invalid;
	string.taps = undefined; // don't save taps data

	if (string.shots.length > 0) { // only save strings with valid shots

		if (typeof global_customPreSaveString === "function") {
			global_customPreSaveString(string);
		}
		
		string.ts = Date.now(); // update timestamp of string
		new_storagejs.write_string(string); // write existing string to file

		if (typeof global_customPostSaveString === "function") {
			global_customPostSaveString(string);
		}

		string.shots = []; // remove shots from memory
		archive[string._id] = string; // put string in archive without shots
		//storage_write_archive(archive); // update archive file
	}

}


function FA_create_string(frame_id, face_id, discard) {
	var start = _performance();
	// console.log('create_string - start (' + _performance(start) + 'ms)');

	var frame = data.frames[frame_id]; // frame
	var string = data.strings[frame.string_id]; // current active string

	// copy geom from frame to string
	string.geom = frame.geom;

	if (!discard) {
		// count valid/invalid shots
		var n_valid = 0;
		var n_invalid = 0;
		var shots_to_save = [];
		for (i in string.shots) {
			var shot = string.shots[i];
			if (!shot) continue; // do not count deleted shots
			shots_to_save.push(shot);
			if (shot.valid) n_valid++; else n_invalid++;
		}

		string.shots = shots_to_save;
		string.n_shots = n_valid;
		string.n_invalid = n_invalid;
		string.taps = undefined; // don't save taps data

		if (string.shots.length > 0) { // only save strings with valid shots
			string.ts = Date.now(); // update timestamp of string
			// console.log('create_string - write_string start (' + _performance(start) + 'ms)');
			new_storagejs.write_string(string); // write existing string to file
			string.shots = []; // remove shots from memory
			// console.log('create_string - archive.setItem start (' + _performance(start) + 'ms)');
			archive.setItem(string._id, string); // put string in archive without shots
			// console.log('create_string - write_archive start (' + _performance(start) + 'ms)');
			//storage_write_archive(archive.getItems()); // update archive file
			// console.log('create_string - write_archive finish (' + _performance(start) + 'ms)');
		}

		delete data.strings[string._id]; // remove string from active data
		// this string will not be deleted from client memory, this is OK
	}

	var str = {}; // create new string with minimum content
	str._id = new_storagejs.storage_get_unique_id();
	str.name = "";
	str.face_id = face_id;
	str.frame_id = frame_id;
	str.shots = [];
	str.ts = Date.now();

	str.autolearn_angles = true;

	// apply string_params from frame
	// console.log('create_string - update_object start (' + _performance(start) + 'ms)');
	if (frame.string_params) update_object(str, frame.string_params);

	crud = {}; // build crud message to apply change to active data
	crud.strings = {};
	crud.strings[str._id] = str; // save new string
	crud.frames = {};
	crud.frames[frame_id] = { string_id: str._id }; // change active string on frame

	apply_crud(crud); // apply crud message

	// console.log('create_string - write_data start (' + _performance(start) + 'ms)');
	storage_write_data(data); // update storage main data
	// console.log('create_string - write_data finish (' + _performance(start) + 'ms)');

	// clear recent shot record
	recent_shot_valid = false;
}

module.exports.clear_all_targets = function () {

	var gStrings = global.data.strings;
	for (var i in gStrings) {
		this.create_string(gStrings[i].frame_id, gStrings[i].face_id);
	}
}

module.exports.create_string = create_string = function (frame_id, face_id) { // used by smtv2-lodi:server.js
		
		var frame = data.frames[frame_id]; // frame
		var string = data.strings[frame.string_id]; // current active string
		var perfStart = performanceLog('create_string: start frame: ' + frame.name);

		var orig_role_scorer = string.role_scorer, orig_role_shooter = string.role_shooter;
		var orig_fclass_target = string.fclass_target;

		// save string to archive
		save_string(string);

		performanceLog('create_string: string saved...', perfStart);

		delete data.strings[string._id]; // remove string from active data
		// this string will not be deleted from client memory, this is OK

		var str = {}; // create new string with minimum content
		str._id = new_storagejs.storage_get_unique_id();
		str.name = "";
		//str.face_id = face_id;
		//str.frame_id = frame_id;
		str.face_id = string.face_id; // attempted bugfix 20-May-2019; ignore caller, duplicate existing
		str.frame_id = string.frame_id; // attempted bugfix 20-May-2019; ignore caller, duplicate existing
		str.shots = [];
		str.validshots = [];
		str.ts = Date.now();
		str.autolearn_angles = true;

		if (data.config.is_mil) { // for mil mode, maintain coach/firer across saves
			str.role_scorer = orig_role_scorer;
			str.role_shooter = orig_role_shooter;
		}
		
		if (data.config.default_bullet_dia) { // 
			str.bullet_dia = data.config.default_bullet_dia;
		}
		
		// preserve F/Iron choice across saves
		str.fclass_target = orig_fclass_target;

		if (typeof global_customCreateStringDefaults === "function") {
			global_customCreateStringDefaults(string, str);
		}

		// apply string_params from frame
		if (frame.string_params) update_object(str, frame.string_params);
		
		crud = {}; // build crud message to apply change to active data
		crud.strings = {};
		crud.strings[str._id] = str; // save new string
		crud.frames = {};
		crud.frames[frame_id] = { string_id: str._id }; // change active string on frame
		
		apply_crud(crud); // apply crud message
		
        performanceLog('create_string: crud applied...', perfStart);
        
		storage_write_all(); // update storage main data
		
        performanceLog('create_string: storage written...', perfStart);

        // clear recent shot record
		recent_shot_valid = false;
	}

function fn_replay_string(str_id, frame_id) { // dev't/testing: re-process/re-calc timing data
	var frame = data.frames[frame_id];
	if (!frame) return;
	_log.debug("replaying string " + str_id + " on frame " + frame.name);

	new_storagejs.storage_read_string_async(str_id, function (err, str_new) {
		if (err) {
			_log.debug("error opening string!");
			return;
		}
		for (var i in str_new.shots) {
			fn_replay_shot(str_new.shots[i],frame_id);
		}
	});
}

// for recalc_shot, use (string) member "idx_internal"

function fn_replay_shot(shot_to_replay, frame_id_or_name) { // dev't/testing: re-process/re-calc timing data
	var frame = data.frames[frame_id_or_name];
	if (!frame) { // try to find by name
		for (var i in data.frames) {
			var f = data.frames[i];
			if (f.name == frame_id_or_name) {
				frame = f;
				// console.log('fn_replay_shot found frame by name',frame_id_or_name);
				break;
			}
		}
	}
//else console.log('fn_replay_shot found frame by id',frame_id_or_name);
	if (!frame) return;
	var frame_id = frame._id;
//console.log('trace fn_replay_shot frame=',frame,'frame_id_or_name=',frame_id_or_name);
//console.log('trace fn_replay_shot frame=',frame,'shot_to_replay=',shot_to_replay);

/* data from a saved string on a G2:
DMC s2= { ts: 1511637052081,
source: 'G2',
x: -121,
y: 109,
v: 587.8,
tempC: 10.2,
num: 333,
mics: [ 5444, 2045, 0, 7170, 10982, -1, -1, -1 ],
mic_mapping: '12345',
valid: true,
value: '9',
score: 9,
shooter_num: 0 }

*/
if (!shot_to_replay.shot) shot_to_replay.shot = { // if raw shot data 'shot' is missing, attempt to re-construct it
		shotnum:shot_to_replay.num,
		// unfortunately this won't allow us to discover the
		// original temperature reading before temp_cal was applied
		G2_temp_rdg: trim_num(shot_to_replay.tempC-frame.geom.temp_cal,2),

//debug:console.log("DMC shot_to_replay=",shot_to_replay),
//debug:console.log("DMC shot_to_replay.shot=",shot_to_replay.shot),
//debug:console.log("DMC shot_to_replay.res=",shot_to_replay.res),
/*
		G2_temp_rdg: (shot_to_replay.res &&  'undefined' != typeof shot_to_replay.res.temp) ? 
		  shot_to_replay.res.temp-frame.geom.temp_cal
		  :
		,
*/
//				counters:mic_mapping.logical_to_phys(shot_to_replay.res.mic_mapping || "12345",shot_to_replay.mics)
		counters:mic_mapping.logical_to_phys(shot_to_replay.mic_mapping || "12345",shot_to_replay.mics),
	};
	shot_to_replay.replayed_shot = true;
	delete shot_to_replay['tempC'];
	delete shot_to_replay['temp_invalid'];
	frameid_emit_shot_direct(frame_id,shot_to_replay);

}

function fn_insert_missed_shot( missingShot ) {
// call as: g_socket.emit("insert_missed_shot", { stringId: cur_string().id(), shotIndex: index, shooter_num: shot.shooter_num});	
	var str = data.strings[missingShot.stringId];
	if (!str) return;

	var crud = {};
	crud.strings = {};
	crud.strings[str._id] = {};
	
	// find the index
	if (str.shots[missingShot.shotIndex]) {
		str.shots.splice(missingShot.shotIndex, 0, {});
	} else {
		str.shots.push({});
	}

	str.shots[missingShot.shotIndex] = { 
		source: "missed_shot",
		ts: Date.now(),
		valid: true,
		x:0, y:0, tempC: 22, v:0, // v == 0 means that client_face.js will not show a "spotting disc"
		shooter_num: missingShot.shooter_num, // miss belongs to same shooter
		value: "0", score: 0,
		res: {
		  theta: 0, phi:0,
		  xy_valid: true
		}
	};
	crud.strings[str._id].shots = str.shots;

	apply_crud(crud);
};


// import targetfaces
function fn_import_targetfaces_new(req, res, next) {
	var file = req.files[ "selected_file" ];
	if ( !file ) {
		res.json( { valid: false, errors: [ "No targetdefinition file found!" ] } );
	}
	_log.debug( "importing target faces from: " + file.originalFilename, file.path );
	var fclass_dualmode_supported = true;
	if (global.is_SOLO) fclass_dualmode_supported = false;

	new_storagejs.new_storage_import_targetfaces( file, fclass_dualmode_supported, function( valid, errors ) {
		_log.debug("/import_targetfaces_new callback ",valid,errors);
		res.json( { valid: valid, errors: errors, msg:errors } );
	} );
}

function fn_import_targetfaces (req, res, next) {
	var file = req.files["selected_file"];
	if (!file) return;
	_log.debug("importing target faces from: " + file.originalFilename, file.path);
	var fclass_dualmode_supported = true;
	if (global.is_SOLO) fclass_dualmode_supported = false;
	new_storagejs.new_storage_import_targetfaces( file, fclass_dualmode_supported, function( valid, errors ) {
		if (!valid) {
			res.json( { valid: valid, errors: errors, msg:errors } );
		} else {
			res.redirect("back");
		}
	});
}

// DMC 70-June-2017: get /archive is not yet used by the client, but we need to implement it soon
// in order to improve fetch performance
//app.get("/archive", function(req, res) {
var get_archive = function(req, res) {
	var user_id = req.query.user_id == "0" ? 0 : req.query.user_id; // string, or undefined, or 0
	var event_id = req.query.event_id === undefined ? 0 : parseInt(req.query.event_id,10);
	var final = req.query.final == 'true';
	var recent = req.query.recent == 'true';
	var only_framename = req.query.only_framename === undefined ? "" : req.query.only_framename;

	_log.debug("GET /archive",user_id, typeof user_id, event_id, typeof event_id, final, typeof final, recent, typeof recent, only_framename, typeof only_framename);

	var start = performanceLog('start GET /archive...');
	var ts_limit = Date.now() - 1000 * 60 * 60 * 24 * 14; // within two weeks from now
	var max_items = 500; // no point returning crazy-lots of records

	var list = [];
	for (i in archive) {
		var str = archive[i];

		if (list.length >= max_items) continue;
		if (recent && str.ts < ts_limit) continue;

		if (event_id >= 0 && (!str.event_data || str.event_data.event_id != event_id)) continue;

		if (final && str.shots_mode > 0) {
			var include = false;
			for (j in str.score) {
				if (str.score[j].count == str.shots_mode) {
					include = true;
					break;
				}
			}
			if (!include) continue;
		}

		if (user_id && str.use_role_shooter) {
			var include = false;
			if (str.role_scorer == user_id) {
				include = true;
			}
			for (j in str.role_shooter) {
				if (str.role_shooter[j] == user_id) {
					include = true;
					break;
				}
			}
			if (!include) continue;
		}

		if ("" != only_framename) { // filter only one targetframe
			var frame = data.frames[str.frame_id];
			if (frame && frame.name != only_framename) continue;
		}

		list.push(str);
	}

	performanceLog('GET /archive... list generated... stringify...', start);
	var str=JSON.stringify(list,null,'  '); // 3295 tuples: 2383-43ms
	performanceLog('GET /archive... list generated... start to send '+list.length+' tuples...', start);
	//res.send(str); // 3295 tuples: 2819-2383ms
	res.end(str); // res.end(str) better than res.send(str)
	performanceLog('finished sending /archive to client...  ' + str.length + ' bytes', start);
}

var _performanceLogEnabled = true;

function performanceLog(logmsg, start) {
    if (!_performanceLogEnabled) return 0;
    if (!start && !logmsg) return _performance();
    if (logmsg && !start) {
        _log.debug('0ms - ' + logmsg);
        return _performance();
    }
    var diff = _performance(start);
    if (logmsg) {
        _log.debug(diff + 'ms - ' + logmsg);
    }
    return diff;
}
module.exports.performanceLog = performanceLog;



// sequence id increments each heartbeat
var sequence_id = 0;

// empty heartbeat
var delta = {};

// TODO: move send_delta in serverjs_handlers, merge with G2 and LODI
// cloned it into G2 2018-11-30

// send heartbeats continously
// check often, send periodically or when changed
var delta_counter = 0;
var delta_changed = false;
var fast_delta_interval = 250; // check every 250ms whether a delta needs to be flushed
var heartbeat_n_fast_deltas = 6; // heartbeat gets sent every 6*250ms = 1500ms regardless of change

var delta_send_timecount = undefined;
var delta_send_t0;
deltasSent = {};
deltasKeys=[];

var connection_limited_to_frameid = function (c/*connection*/) {
	if (!c || !c.connection_info) return undefined;
	var userid = c.connection_info.recentuser;
	var user = data.users[userid];
	// console.log("connection_limited_to_frame, userid",userid,"=",user)
	if (user) {
		if (user.name == "admin") return undefined;
		if (user.is_range_officer) return undefined;
	}
	if (c.connection_info.showallframes) { // opportunity for an ordinary user to get all shots
		return undefined;
	}
	// are we connected to a particular frame?
	var frameid = c.connection_info.recentframe;
	// console.log("connection_limited_to_frame, userid",userid,"frameid=",frame);
	return frameid;
}

var connection_recentframe_sema_check = function (c/*connection*/) {
	if (c && c.connection_info && c.connection_info.connection_recentframe_sema) {
		return true;
	}
	return false;
}

var send_delta = function () {
	// only send if needed
	// if (!delta_changed && ++delta_counter < 6) return;
	// if (++delta_counter < 12) return;
	var start = _performance();
	var time_since_last_delta = _performance(delta_send_timecount);
	delta.interval = heartbeat_n_fast_deltas * fast_delta_interval; // normally 1500ms
	var send_delta_now = false;
	if (time_since_last_delta > delta.interval) {
		send_delta_now = true;
	}
	if (delta_changed) send_delta_now = true;
	if (++delta_counter >= 16) send_delta_now = true;

	// console.log('Tobias paranoid', send_delta_now, time_since_last_delta,'alt',Date.now()-delta_send_t0);
	if (!send_delta_now) {
		// console.log("send_delta, NOT sending, time since last =",time_since_last_delta)
//		setTimeout(send_delta, 250);
		return;
	}
	// console.log("send_delta, sending, time since last =",time_since_last_delta)
	
	delta.seq = ++sequence_id;
	delta.ts = Date.now();
	// if (data.config.wastey) {
	// 	var i, b=[1];
	// 	for(i = 0; i < data.config.wastey; i++) b = [b,b];
	// 	delta.wastey = b;
	// 	// delta.wastey = Array(data.config.wastey).join('x');
	// }
	if (data.config.experimental_features && data.config.experimental_features.highperf) {
		var sockets = io.of("/").sockets;
		// console.log("DMC highperf len=",sockets.length, "size=", sockets.size, "clientsCount=",io.engine.clientsCount);
		var restrict_to_frameid = undefined;
		// save:
		var allstrings = delta.strings; // could be undefined
		var alltaps = delta.taps; // could be undefined
		// also: string.taps
		for (var i in sockets) {
			var socket = sockets[i];
			var socketid = sockets[i].id;
			var c = data.connections[socketid];
			restrict_to_frameid = connection_limited_to_frameid(c);
			var restore_allshots = false;
			var string = undefined,string_saved_shots = undefined;
			if (restrict_to_frameid) {
				var frame = data.frames[restrict_to_frameid];
				var stringid = frame && frame.string_id;
				console.log("frame ",frame && frame.name,"stringid=",stringid);
				restore_allshots = true;
				delta.strings = {};
				if (stringid) {
					string = delta.strings[stringid] = allstrings && allstrings[stringid];
					string_saved_shots = string && string.shots; // could be undefined
					if (string) {
						string.shots = undefined;
						// validshots is what counts
					}
				}
				// TODO: trim out allshots
			} else { // admin etc
				delta.strings = allstrings;
			}
			// console.log("blah i",i,"c",c,delta);
			// TODO: perhaps restrict data
			socket.emit("delta", delta);
			if (restore_allshots) { // restore
				if (string_saved_shots) {
					string.shots = string_saved_shots;
				}
			}
		}
		// restore:
		delta.strings = allstrings;
	} else {
		// console.log("NOT higherf")
		// old way - slower, but works:
		io.emit("delta", delta);
	}

	// // old way - slower, but works:
	// io.emit("delta", delta);
	var et_delta= _performance(start);
	var start = _performance();
	var len=JSON.stringify(delta).length;
	var et_stringify = _performance(start);
	// console.log('send_delta (' + et_delta + 'ms)',"delta size=",len,'stringify time=',et_stringify);

	// new way: more promising, but we need to work ou the kinks
	// io.emit("delta_no", { // very short message, only announcing latest delta #
	// 	seq: delta.seq,
	// 	ts: delta.ts,
	// 	interval: delta.interval	
	// });


	var et_delta= _performance(start);
	// console.log('send_delta_no (' + et_delta + 'ms)');

	// var start = _performance();
	// if (0 && typeof global.gc == 'function'){
	// 	global.gc();
	// 	console.log('send_delta, global.c(), time= ' + _performance(start) + 'ms');
	// }
	
	// console.log('send_delta (' + _performance(start) + 'ms)',"delta size=",(JSON.stringify(delta)).length);
	// console.log('send_delta (' + _performance(start) + 'ms)');
	// var a;
	//console.log("delta size=",(a=JSON.stringify(delta)).length/*,JSON.stringify(delta,null," "),"len=",a.length*/);
	// console.log('send_delta - 2 (' + _performance(start) + 'ms)');

	// clear delta for next heartbeat
if (1) {
	last_delta_sent = delta;
	deltasSent[delta.seq] = JSON.stringify(delta);
	deltasKeys.push(delta.seq);
	// too big?
	var len = deltasKeys.length;
	var maxlen = 20;
	if (len > maxlen) {
		var i;
		for (i = 0; i < len-maxlen; i++) {
			var keyToRemove = deltasKeys.shift();
			delete deltasSent[keyToRemove];
		}
	}
}
	//
	delta = {};
	delta_changed = false;
	delta_counter = 0;
	delta_send_timecount = _performance();
	delta_send_t0 = Date.now();
	// setTimeout(send_delta, 250);
}


// start heartbeat timer
delta_send_timecount = _performance();
delta_send_t0 = Date.now();
setInterval(send_delta, fast_delta_interval);


var cached_data_string = "";
var cached_data_id = -1;

global.data_changed = false;

// receive and apply changes to data
// does not automatically update storage
function apply_crud(msg) {
	_log.trace("crud:\n", msg);
	update_object(data, msg); // apply to local data
	update_object(delta, msg, {create_delta:true}); // apply to next heartbeat; special flag to propagate rather than processg_special_delete_obj
	delta_changed = true; // send next heartbeat
	data_changed = true; // DMC
	cached_data_string = ""; // data changed; clear cache
}
global.apply_crud = apply_crud; // klugey, yup, but want to access it from require()'d functions, namely lobots_control.*

// broadcast full data to everyone
// only used when heartbeat will not suffice (deleting frames/faces)
function broadcast_data() {

	// we are calling broadcast because something has changed outside of a crud, so...
	cached_data_string = ""; // clear cache
	
	// anything in delta is now out of date
	delta = {};
	delta_changed = false;
	delta_counter = 0;
	
	// DMC: the assumption is made that broadcast_data has almost certainly been
	// called because a change has been made to the data
	data_changed = true;
	
	sequence_id = 0; // reset just to be safe
	
	// client will request new data on its own time
	io.emit("request_data");
	
}

global.broadcast_data = broadcast_data; // klugey, yup,

