
var toolbar_review = new function() {
	
	this.report_strings = []; // strings currently in the report list
	this.display_admin = false; // whether to show invalid shots
	
	this.string_opened_last = null; // most recently opened string
	
	// whether archive search filters are on
	this.filter_user = true;
	this.filter_recent = true;
	this.filter_event = true;
	this.filter_final = false; // DMC; true is a worse default option
	this.filter_curtgt_only = false; // DMC

	this.init = function() {
		this.icon = $("<img src='images/strings.png' class='toolbar_icon'>");
		this.folder_icon = $("<img src='images/strings.png' class='toolbar_icon inline_icon'>").css("padding", "3px");
		this.admin_icon = $("<img src='images/admin.png' class='toolbar_icon inline_icon'>").css("padding", "3px");
		this.plot_icon = $("<img src='images/plotting.png' class='toolbar_icon inline_icon'>").css("padding", "3px");
        this.mil_icon = $("<img src='images/admin.png' class='toolbar_icon inline_icon'>").css("padding", "3px");
		this.print_icon = $("<img src='images/print.png' class='toolbar_icon inline_icon'>").css("padding", "3px");
	}
	
	// whether icon should be shown on main toolbar
	// if using assigned shooters, must be logged in to review strings
	this.ready = function() {
		if (toolbar_military.is_mil()) return true;
		return !g_config.use_role_shooter || cur_user() != null;
	}
	
	// try to open this panel (return false to cancel)
	this.show = function() {
		this.display_admin = false;
		plotting.enable_view = false;
		if (live_mode()) { // coming here from live mode
			this.previous_frame = cur_string().frame(); // remember frame
			if (numberof(g_archive) == 0) this.refresh(); // auto refresh first time
		}
		return true;
	}
	
	// try to close this panel (return false to cancel)
	this.hide = function() {
		plotting.enable_view = false;
		// try to select current string on the same frame
		if (this.previous_frame && this.previous_frame.string()) {
			select_string(this.previous_frame.string().id());
		} else {
			select_string(null); // select any valid string
		}
		return true;
	}
	
	this.refresh = function() {
		var user_id = 0;
		if (toolbar_review.filter_user && cur_user() && cur_user().data.name != "admin")
			user_id = cur_user().id();
			
		// special restriction: if using assigned shooters, and user is not ro or admin, force current user filter
		if (!user_id && cur_user() != null && g_config.use_role_shooter && !ro_check() && !admin_check()) {
			user_id = cur_user().id();
		}

		var event_id = -1;
		if (toolbar_review.filter_event && cur_event())
			event_id = cur_event().id;
		var final = toolbar_review.filter_final;
		var recent = toolbar_review.filter_recent;
		var old_way = false;
		if (old_way) {
			g_socket.emit("get_archive", user_id, event_id, final, recent
				,toolbar_review.filter_curtgt_only ? cur_string().frame().data.name : ""
			); // DMC todo - add different filtering options
		} else {
			request_get_archive(user_id, event_id, final, recent
				,toolbar_review.filter_curtgt_only ? cur_string().frame().data.name : ""
			); // DMC todo - add different filtering options
		}
	}

	// display the panel in the supplied div
	this.draw = function(div) {

		if (!live_mode()) {
			
			this.string_opened_last = cur_string().id();
			
			// folder icon to open list again
			this.folder_icon.mousedown(function() {
				log(toolbar.back());
				log(toolbar.set_panel(toolbar_review));
				update();
			});
			div.append(this.folder_icon);

			// string name
			toolbar.new_text(div, cur_string().data.name);
							
			this.plot_icon.mousedown(function() {
				plotting.enable_view = !plotting.enable_view;
				plotting.enable_controls = false;
				shotbubble.enabled = false; // because there is no shotbubble without plotting
				update();
			});
			div.append(this.plot_icon);
			
			if (plotting.enable_view) this.plot_icon.css("border", "2px solid");
			else this.plot_icon.css("border", "0");

			if (window.print) {
				this.print_icon.mousedown(function() {
					print_screen();
				});
				div.append(this.print_icon);
			}

			div.append("<br>");
			
/*
			div.append("<br>");
			toolbar.new_text(div, S("REVIEW_REPORT"));

			// check if this string is already in the pending report list
			var found = false;
			for (var i in this.report_strings) {
				var str = this.report_strings[i];
				if (str.id == cur_string().id() && str.shooter == cur_shooter) {
					found = true;
					break;
				}
			}
			if (!found) {
				toolbar.new_button(div, S("REVIEW_REPORT_ADD"), function() {
					toolbar_review.report_strings.push({
						id: cur_string().id(),
						shooter: cur_shooter
					});
					update();
				});
			}

			// download report button
			if (this.report_strings.length > 0) {
				toolbar.new_button(div, S("REVIEW_REPORT_DOWNLOAD", this.report_strings.length), function() {
					g_socket.emit("download_report", toolbar_review.report_strings);
					toolbar_review.report_strings = [];
				});
			}
*/
			// admin: display invalid shots
			if (admin_check() && cur_string().data.n_invalid > 0) {
				div.append("<br>");
				toolbar.new_checkbox(div, S("REVIEW_DISPLAYINVALID", cur_string().data.n_invalid), function() {
					toolbar_review.display_admin = !toolbar_review.display_admin;
					update();
				}, toolbar_review.display_admin);
			}
			
			if (admin_check() || ro_check()) {
				
				div.append("<br>");
				toolbar.new_text(div, "Restore string on:");
				

				// determine frames to allow restore string
				var frames_to_display = [];
				for (var i in g_frames) {
					var frame = g_frames[i];
					if (!frame_is_valid_to_select(frame)) continue; // check against admin/frame status
					frames_to_display.push(i);
				}

				frames_to_display.sort(function(a,b) {
					if (g_frames[a].data.name == g_frames[b].data.name) return 0;
					if (g_frames[a].data.name == "new") return -1;
					if (g_frames[b].data.name == "new") return 1;
					return g_frames[a].data.name.toUpperCase().localeCompare(g_frames[b].data.name.toUpperCase());
				});
				
				var list = [];
				list.push({value: 0, text: "Choose a target..."});
				for (i in frames_to_display) {
					var id = frames_to_display[i];
					var frame = g_frames[id];
					list.push({value: id, text: frame.data.name});
				}
				
				toolbar.new_dropdown(div, list, function(item) {
					var frame = g_frames[item.val()];
					if (!frame) return;
					
					if (frame.string().data.shots.length > 0) {
						var res = confirm("string has shots! replace?");
						if (!res) {
							update(); // to re-select 'choose a frame'
							return;
						}
					}
					
					var str_id = cur_string().id();
					
					toolbar.back();
					targetrow.selectFrame(frame.id());
					update();
					
					g_socket.emit("revive_string", str_id, frame.id());
					
				}, 0);
				
				// DMC
				
				if (true || g_config.is_JRPL) {

//					toolbar.new_text(div, "REPLAY string on:");
				
					// determine frames to allow restore string
					var frames_to_display = [];
					for (var i in g_frames) {
						var frame = g_frames[i];
						if (!frame_is_valid_to_select(frame)) continue; // check against admin/frame status
						frames_to_display.push(i);
					}

					frames_to_display.sort(function(a,b) {
						if (g_frames[a].data.name == g_frames[b].data.name) return 0;
						if (g_frames[a].data.name == "new") return -1;
						if (g_frames[b].data.name == "new") return 1;
						return g_frames[a].data.name.toUpperCase().localeCompare(g_frames[b].data.name.toUpperCase());
					});
				
					var list = [];
					list.push({value: 0, text: "Replay string on..."});
					for (i in frames_to_display) {
						var id = frames_to_display[i];
						var frame = g_frames[id];
						list.push({value: id, text: frame.data.name});
					}
				
					toolbar.new_dropdown(div, list, function(item) {
						var frame = g_frames[item.val()];
						if (!frame) return;
					
						var str_id = cur_string().id();

						var answer = confirm("CONFIRM: replay on " + frame.data.name+ " ?");
						if (!answer) return;
					
						toolbar.back();
/*						targetrow.selectFrame(frame.id());
						update();
*/
					
						g_socket.emit("replay_string", str_id, frame.id());
					
					}, 0);
				}
				
			}

		// 'choose a string' table is open
		} else {

			if (cur_user() != null && cur_user().data.name != "admin") {
				var item = toolbar.new_checkbox(div, "Current user", function() {
					toolbar_review.filter_user = !toolbar_review.filter_user;
				}, toolbar_review.filter_user);
				
				// special case: force the checkbox
				if (g_config.use_role_shooter && !ro_check()) {
					item.attr("disabled", "true");
					item.attr("checked", "true");
				}
			}
			
			if (cur_event()) {
				toolbar.new_checkbox(div, "Current event", function() {
					toolbar_review.filter_event = !toolbar_review.filter_event;
				}, toolbar_review.filter_event);
			}

			toolbar.new_checkbox(div, "Only final", function() {
				toolbar_review.filter_final = !toolbar_review.filter_final;
			}, toolbar_review.filter_final);
			
			toolbar.new_checkbox(div, "Recent 2 weeks", function() {
				toolbar_review.filter_recent = !toolbar_review.filter_recent;
			}, toolbar_review.filter_recent);

			toolbar.new_checkbox(div, "Only "+cur_string().frame().data.name, function() { // DMC
				toolbar_review.filter_curtgt_only = !toolbar_review.filter_curtgt_only;
			}, toolbar_review.filter_curtgt_only);
		
			toolbar.new_button(div, "Refresh", function() {
				toolbar_review.refresh();
			});

			// build local array filtered by something
			var strings = [];
	
			for (var i in g_archive) {
				var str = g_archive[i];
	
				// *** filter ***
	
				strings.push(str);
			}
	
			// sort newest first
			strings.sort(function(a,b) {
			    if (a.ts > b.ts) return -1;
			    if (a.ts < b.ts) return 1;
			    return 0;
			});
	
			function string_date_text(string) {
				var date_string = new Date(string.ts);
				var date_now = new Date();
				var date_text = "";
				if ((new Date()).toDateString() == date_string.toDateString()) { // today
					date_text = date_string.toLocaleTimeString();
				} else { // "month day"
					var date_words = date_string.toDateString().split(" ");
					date_text = date_words[1] + " " + date_words[2] + " " + date_string.toLocaleTimeString();
				}
	
				return date_text;
			}

			// report # strings
			div.append('('+strings.length+')');
			div.append("<br>");

			var divString = $("<div>");
			divString.css("width", maintarget.div.width() * 0.90);
			divString.css("height", maintarget.div.height() * 0.80);
			divString.css("overflow-y", "scroll");
			divString.css("margin", "10px");

			var table = $("<table id='tblStringReview'>");
			table.css("width", "100%");

			// allow touch scrolling only when scrolling is possible
			// to prevent scrolling the entire page itself
			table.bind("touchstart", function(e) {
				toolbar.table_touch_start = e.originalEvent.touches[0].pageY;
				e.originalEvent.allow_touchstart = true;
			});

			table.bind("touchmove", function(e) {
				var delta = e.originalEvent.touches[0].pageY - toolbar.table_touch_start;
				if (delta > 0 && divString[0].scrollTop == 0) return;
				if (delta < 0 && divString[0].scrollTop == divString[0].scrollHeight - divString[0].clientHeight) return;
				e.originalEvent.allow_touchmove = true;
			});

			var tr_selected = null;

			table.append("<tr><th>" + S("REVIEW_TABLE_DATE") + "</th><th>" + S("REVIEW_TABLE_COUNT") + "</th><th>" + S("REVIEW_TABLE_NAME") + "</th><th>" + S("REVIEW_TABLE_TARGET") + "</th></tr>");

			if (strings.length > 0) {

				var found_cur = false;
//				g_frames[undefined] = {data:{name:'-undefined-'}};
//				g_faces[undefined] = {data:{name:'-undefined-'}};
				for (var i in strings) {
					var string = strings[i];
					
					if (!g_frames[string.frame_id]) {
//						g_frames[string.frame_id] = g_frames[undefined];
					}
					if (!g_faces[string.face_id]) {
//						g_faces[string.face_id = g_faces[undefined];
					}

					// skip over strings using deleted frame or face
					if (!g_frames[string.frame_id] || !g_faces[string.face_id]) {
						continue;
					}

					var tr = $("<tr>");
					tr[0].arg = string._id;
					tr.mousedown(function() {
						g_socket.emit("get_string", {string_id: this.arg});
						update();
					});

					var shot_text = "";
					shot_text = string.n_shots;
					if (admin_check() && string.n_invalid) {
						shot_text += " (" + string.n_invalid + ")";
					}
					shot_text += " shots";
					
					var string_text = "";
					if (typeof string.name == 'string') string_text = string.name;
					if (string_text == "") string_text = "-";
					for (var j = 0; j < string.shooter_count; j++) {
						if (string.role_shooter && string.role_shooter[j]) {
							var user = g_users[string.role_shooter[j]];
							if (user) {
								// realname could be undefined
								string_text += ", " + (user.data.realname || user.data.name);
							}
						}
						if (string.score && string.score[j]) {
							string_text += " (" + string.score[j].string + ")";
						}
					}
					
					var event_text = "";
					if (string.event_data) {
						shot_text += " "+g_frames[string.frame_id].data.name; // DMC - target name - fit it in with # shots" column
						event_text += string.event_data.event_str + "<br>" + string.event_data.match_str + "<br>" + string.event_data.stage_str;
					} else {
						event_text += g_frames[string.frame_id].data.name + " (" + g_faces[string.face_id].data.name + ")";
					}
					

					tr.append("<td>" + string_date_text(string) + "</td><td>" + shot_text + "</td><td>" + string_text + "</td><td>" + event_text + "</td>");

					table.append(tr);
					
					if (string._id == this.string_opened_last) {
						found_cur = true;
						tr.addClass("selected");
						tr_selected = tr;
					}

				}

			}

			divString.append(table);
			div.append(divString);

			if (tr_selected) {
				divString.scrollTop(tr_selected[0].offsetTop - divString.height() / 2);
			}
			
		}

	}
		
	
}

