/*******************************************************************************
 new_storage.js: common "storage" routines, flatfiles <--> internal
 
 Usage:

 $Author: dmc $
 $Revision: 3226 $
 $Date: 2024-01-19 15:19:31 -0400 (Fri, 19 Jan 2024) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/new_storage.js $
*******************************************************************************/

//function performanceLog(logmsg, start) {} // short-circuit for now
//function log() {} // eval(fs.readFileSync("./smt.common.lib/public/javascripts/utils.js")+""); // common utils

var log4js = require( "log4js" );
var _log = log4js.getLogger( "new_storage.js" );
/*
console.log("_log=",_log);
console.log("_log.level=",_log.level);
console.log("setLevel DEBUG");
_log.setLevel("DEBUG");
console.log("_log.level=",_log.level);
//_log.level = 'debug';
//console.log("set to debug; test!");
_log.info("hi there - INFO!");
*/

var _performance = require( "./lib/utils/performance.js" );

var exec = require("child_process").exec;
var zlibjs = require("zlib"); // use native... it seems to work.  "zlibjs" is sometimes broken
if (!zlibjs || !zlibjs.inflateSync) zlibjs = require("zlibjs"); // nodejs v0.10.22 (S10) doesn't have inflateSync

var fs = require("fs"); // file system
var os = require("os");

function storage_update_diskspace() {
	var df_cmd = "";
// linux desktop ubuntu df . -kP | cut -c37-46
// Available
//      3000
//
// Ubuntu 15.04 df . -kP | cut -c48-57
// Available
// 200633256
//
// OpenWRT x86_64
//root@SMT_S50_server:~# df . -kP | cut -c41-50
// Available
//    228308
//
// OpenWRT RPi-4
//root@SMT_S7_server:~# uname -a
//Linux SMT_S7_server 4.19.91 #0 SMP Mon Dec 30 18:32:10 2019 aarch64 GNU/Linux
//root@SMT_S7_server:~# df . -kP | cut -c41-50
// Available
//    376300
//
// SOLO:
//root@SMT_SOLO:~# uname -a
//Linux SMT_SOLO 3.18.109 #55 Thu May 28 14:40:05 ADT 2020 mips GNU/Linux
//root@SMT_SOLO:~# df . -kP | cut -c41-50
// Available
//      6492


//	if (os.arch()=='x64' && 'v0' == process.version.substr(0,process.version.search("\\."))) {
	if (os.arch()=='x64' && '0.10.25' == process.versions.node) {
		df_cmd = 'df . -kP | cut -c48-57 | tail -1';
	} else if (os.arch()=='arm64' || os.arch()=='mips' || os.arch()=='x64') { // BusyBox v1.31.1
		df_cmd = 'df . -kP | cut -c41-50 | tail -1';
	}

/*
APU1D BusyBox v1.31.1 snapshot July-2020 (post-19.07)
root@SMT_S50_server:/# df . -kP | cut -c41-50
 Available
    357564
*/ 
	// https://stackoverflow.com/questions/44647778/how-to-run-shell-script-file-using-nodejs
	if (df_cmd) {
		var yourscript = exec(df_cmd, function (error, stdout, stderr) {
            if (error !== null) return;
			var free_shots = Math.floor(parseInt(stdout) * 1024 / 40);
			apply_crud({config: {diskspace: free_shots}});
        });
    }
}


var archive2;
module.exports.archive2 = archive2 = {};

module.exports.rstat_ready = false; // will be set after all strings are read (which could take some time)

module.exports.storage_write_rstat = function (archive) {
	var rstat_data = [];
	var cnt_saved_true = 0;
	var cnt_saved = 0;
	for (var i in archive) {
		var s = archive[i];
		if ('object' == typeof s && 'undefined' != typeof s.rstat_saved) {
			rstat_data.push( {idx: i, rstat_saved: !!s.rstat_saved} );
			cnt_saved++;
			if (!!s.rstat_saved) cnt_saved_true++;
		}
	}
	if (rstat_data.length == 0) {
		console.log("new_storagejs.storage_write_rstat() length=0, not writing rstat.txt");
		fs.unlinkSync("data/rstat.txt");
		return;
	}
	console.log("new_storagejs.storage_write_rstat()...",cnt_saved,"strings with rstat_save, #true=",cnt_saved_true);
	fs.writeFile("data/rstat.txt", JSON.stringify(rstat_data, null, "\t"), function(error) {});
}

function read_strings_archive2(string_filenames) {
	if (!string_filenames || string_filenames.length == 0) {
		_log.info("read_strings_archive2 done, archive2.length ==",Object.keys(archive2).length
			,"archive.length ==",Object.keys(archive).length);
		console.log("read_strings_archive2 done, archive2.length ==",Object.keys(archive2).length
			,"archive.length ==",Object.keys(archive).length);
		// now merge in rstat.txt, if it exists:
		var rstat_data;
		var crud = {config: {}};
		try {
			var rstat_data = JSON.parse( fs.readFileSync( "data/rstat.txt" ) );
			crud.config = {is_rstat: true};
			console.log("rstat_data read",crud);
		} catch(e) {
			crud.config = {is_rstat: false};
			console.log("rstat_data not found",crud);
		}
		apply_crud(crud); // automatically show/hide ADMIN->RSTAT button, based on presence of "rstat.txt"
		
		var nset = 0, nlost = 0;
		if (rstat_data) {
			var i;
			for (i = 0; i < rstat_data.length; i++) {
				var rr = rstat_data[i];
				var idx = rr.idx;
				var s = archive[idx];
				if (s) {
					s.rstat_saved = rr.rstat_saved;
					nset++;
				} else {
					nlost++;
				}
			}
		}
		console.log("rstat data nset=",nset, "n_lost=",nlost);
		module.exports.rstat_ready = true;
		return; // no more tail calls
	}

	var fname = string_filenames.shift();
	
	storage_read_string_fname_async(fname, function (err, parsed_string) {
		if (!err) {
			if (parsed_string) {
				parsed_string.shots = [];
				parsed_string.taps = undefined;
				parsed_string.geom = undefined;
				if (parsed_string._id) {
					archive2[parsed_string._id] = parsed_string;
				}
			}
		} else { // error
			_log.error("DMC read_strings_archive2 fname=",fname," err=",err);
		}
		setImmediate(read_strings_archive2, string_filenames); // tail-recurse, but don't hog CPU
	});
}

function build_archive2() {
    var start = _performance();
	_log.debug("build_archive2() start");
	fs.readdir("./data",function(err, files) {
		var string_filenames = [];
		if (err) {
			_log.error("build_archive2() readdir(data) error = ",err);
			throw err;
		}
		if (!err) {
			files.forEach(function(fname) {
				var m;
				if (m=fname.match(/^string-\d+\.(txt)|(zip)$/)) {
					//console.log("m=",m);
					string_filenames.push(fname);
				}
			});
			string_filenames.sort().reverse(); // process file names in reverse order
			read_strings_archive2(string_filenames); // read one at a time
		}
		_log.debug("build_archive2() setup done");
	});
}


module.exports.new_storage_init = function ( cb ) {

	exec( "./storage_init.sh", function( err, stdout, stderr ) { // OK if this script does not exist
		// console.log("dmc storage_init.sh err=",err,"stdout=",stdout);
		if ( typeof cb === "function" ) {
			cb( err );
		}
		// now fire off the "dynamically build archive2"
		build_archive2();
	} );
}

function read_parse_sync_and_catch(full_filename) { // can throw; make a copy if file corrupted
	var file_contents, parsed_file_contents;
	try {
		parsed_file_contents = JSON.parse( file_contents = fs.readFileSync( full_filename ) );
	} catch(e) {
		console.log("read_parse_sync_and_catch() Error reading/parsing",full_filename,"...saving copy of corrupt data...  Error=",e);
		fs.writeFileSync( full_filename+".corrupted", file_contents ); // for SMT post-mortem
		throw e;  // pass on to outer handler
	}
	return parsed_file_contents;
}


function read_file_or_backup_Sync(dirname,filename) { // can throw
	var file_contents;
	var full_filename = dirname + "/" + filename;
	try {
		file_contents = read_parse_sync_and_catch(full_filename);
	} catch(e) {
		console.log("Error reading/parsing",full_filename,"...trying backup...");
		full_filename += ".bak";
//		file_contents = JSON.parse( fs.readFileSync( full_filename ) ); // we don't catch this one
		file_contents = read_parse_sync_and_catch(full_filename);
	}
	console.log("read_file_or_backup_Sync() success reading ",full_filename);
	return file_contents;
}

function read_data_faces_dir_Sync(dirname) { // can throw
	var data = read_file_or_backup_Sync(dirname,"data.txt");
	if ( data.faces === undefined ) {
		var faces = read_file_or_backup_Sync(dirname,"faces.txt");
		data.faces = faces;
	} else { // transition to a separate faces.txt..
		storage_write_data( data );
		storage_write_faces( data );
	}
	return data;
}

function import_string_to_fn(face, fn_name_str) { // return undefined, or a parsed functino
	if ('string' != typeof face[fn_name_str]) {
		return undefined;
	}
	var eval_rc = undefined;
	try {
		var evalstr = 'var zz_checkstuff = ' + face[fn_name_str] + ' ; zz_checkstuff';
		eval_rc = eval(evalstr);
	} catch (e) {
		// console.log("DMC eval catch, e=",e)
	}
	console.log("DMC eval done, fn_name_str=",fn_name_str," eval_rc=", eval_rc)
	if ('function' == typeof eval_rc) {
		if (eval_rc.toString() == face[fn_name_str]) {
			console.log("DMC all is well with the world, face.valn.toString() == face.valfn_as_string",fn_name_str,face[fn_name_str] )
		} else {
			console.log("DMC the world is decidedly wrong**** , face.valn.toString() != face.valfn_as_string",fn_name_str)
		}
		return eval_rc;
	}
	return undefined;
}

function fixup_valfn(face) {
	if ('object' != typeof face) {
		// console.log("NOPE NOPE NOPE!!, typeof face=",typeof face)
		return;
	}
	face.hookfn_calc_string_score = import_string_to_fn(face,"hookfn_calc_string_scorer_as_string");
	face.hookfn_recv_shot_assign_shooter = import_string_to_fn(face,"hookfn_recv_shot_assign_shooter_as_string");
}


module.exports.storageReadDataSync =
// read frames, faces, and active strings for initial startup
function ( ) {
	var data = undefined;
	try {
        var start = _performance();
		try {
			data = read_data_faces_dir_Sync("data"); // will try .bak if needed; can throw
		} catch (e) {
			// trouble; all we can do is reset to initial-data
			console.log("storageReadDataSync reading principal/backup, reverting to initial-data..., Error=",e);
			try {
				data = read_data_faces_dir_Sync("initial-data");
			} catch (e) {
				console.log("storageReadDataSync Even reverting FAILED!!!, Error=",e);
				// REAL trouble
			}
		}

		for ( var i in data.strings ) {
			var str = data.strings[ i ];
			if (
				!data.frames[ str.frame_id ] ||
				!data.faces[ str.face_id ] || // bad if no face ID
				data.frames[ str.frame_id ].string_id != i // bugfix: not operator !== please!!
			) {
				_log.error( "Deleting invalid active string!", i );
				console.log( "storageReadDataSync deleting invalid active string!", i );
				delete data.strings[ i ];
			}
		}

		// check that every frame has a valid string and create a new one if needed...
		for ( var i in data.frames ) {
			var frame = data.frames[ i ];
			if ( frame.string_id !== undefined && data.strings[ frame.string_id ] === undefined ) {

				// create a new empty string for the frame and use the first face as default...
				var str = module.exports.createNewStringByFrame( data.frames[ i ], data.faces[ Object.keys( data.faces )[ 0 ] ] );
				str._id = frame.string_id;
				data.strings[ frame.string_id ] = str;
				_log.error( "Create a empty string ( " + frame.string_id + " ) for frame '" + frame.name + "'" );
			}
		}
		
		// make sure every target face is named, otherwise delete it
		for ( var i in data.faces ) {
			if (undefined == data.faces[i].name) {
				delete data.faces[i];
			}
			// DMC 18 Jan 2024
			var face = data.faces[i];
			fixup_valfn(face);
			// if ('string' == typeof face.valfn_as_string) {
			// 	console.log("DMC Blah",face.valfn_as_string);
			// 	var eval_rc = undefined;
			// 	try {
			// 		var evalstr = 'var zz_checkstuff = ' + face.valfn_as_string + ' ; zz_checkstuff';
			// 		eval_rc = eval(evalstr);
			// 	} catch(e) {
			// 		// console.log("DMC eval catch, e=",e)
			// 	}
			// 	console.log("DMC eval done, eval_rc=",eval_rc)
			// 	face.valfn = undefined;
			// 	if ('function' == typeof eval_rc) {
			// 		face.valfn = eval_rc;
			// 		if (face.valfn.toString() == face.valfn_as_string) {
			// 			console.log("DMC all is well with the world, face.valn.toString() == face.valfn_as_string")
			// 		} else {
			// 			console.log("DMC the world is decidedly wrong**** , face.valn.toString() != face.valfn_as_string")
			// 		}
			// 	}
			// }
		}

		if (!data.config || !data.config.language) data.config = {language: "en"};
		data.config.device_type = (is_G2)?("G2"):
			((is_R5)?("R5"):
				((is_S50)?("S50"):
					("S10")
				)
			);
		if (!data.users) data.users = {
			"1": {
				"_id": 1,
				"name": "admin",
				//"password": "becareful", // default: no password
				"fclass_target": true,
				"bullet_dia": 7.82
			},
			"2": {
				"_id": 2,
				"name": "ro",
				//"password": "becareful", // default: no password
				"fclass_target": true,
				"bullet_dia": 7.82
			},
		};
		if (!data.events) data.events = {};

		if ('function' == typeof storage_migrate) {
			storage_migrate( data ); // changes from older schemas/releases
		}

		console.log( "storageReadDataSync (" + _performance( start ) + "ms)" );

	} catch ( e ) {
		_log.error( "error reading data from file", e );
	}
	// return what???
	return data;
}

// will delete and replace EVERYTHING in data folder
// will accept all file types
// will also accept and process software update files
module.exports.storage_restore_all_data =
function (filepath, cb) {
	console.log("storage_restore_all_data",filepath);
	exec( "sh smt.common.lib/restore_backup.sh '"+filepath+"'", function( err, stdout, stderr ) { // OK if this script does not exist
		console.log("restore_backup.sh err=",err);
		console.log("stdout=",stdout);
		console.log("stderr=",stderr);
		if ( typeof cb === "function" ) {
			cb( err, stdout, stderr);
		}
	} );
}


// remove symlink but do not crash if the file doesn't exist
function rm_symlinkSync(name) {
	try { fs.unlinkSync(name); } catch(e) {}; // ignore errors
//	fs.existsSync(name) && fs.unlinkSync(name); // this gets done if symlink target exists
//	fs.lstatSync(name) && fs.unlinkSync(name); // this works if symlink target does not exist
}

// produce a "report.csv" file, for several string ids
module.exports.storage_write_report = function (report_strings) {
/* is broke:
console.log("storage_write_report _log.level=",_log.level);
_log.setLevel("DEBUG");
_log.info("error - error");
_log.warn("test - warn");
_log.info("test - info");
_log.debug("test - debug");
_log.trace("test - trace");

console.log("2-storage_write_report _log.level=",_log.level);
	_log.info("storage_write_report()",report_strings);
*/
	//console.log("DMC debug  - info","storage_write_report()",report_strings);
	
	// var filename = "public/report.csv";
	var linkToTmpPath = "public/reports";
	var filename = linkToTmpPath + "/report.csv";
	var tmpPath = "/tmp/shot-reports";

	if (fs.existsSync("/tmp/")) { // there is /tmp, so use it instead
		rm_symlinkSync(linkToTmpPath);
		fs.existsSync(tmpPath) || fs.mkdirSync(tmpPath);
		fs.symlinkSync(tmpPath,linkToTmpPath);
	}

	var speed_unit = "(m/s)", speed_mpier=1, speed_ndigs=1;
	if (data.config.option_speed_unit == 'fps') { // off,ms,fps 
		speed_unit = "(fps)"; speed_mpier = 3.28; speed_ndigs=0;
	}
	var xy_unit = "(mm)", xy_mpier=1, xy_ndigs=1;
	if (data.config.option_coords == '"') { // mm,"
		xy_unit = '(")'; xy_mpier = 1/25.4; xy_ndigs=2;
	}

	var out_str = '';

	// fs.writeSync(out, "string,shooter,frame,distance,date,score,moa_x,moa_y,scope_x,scope_y,adj_x,adj_y,v,adj_y_avg,adj_y_sd,v_avg,v_sd\n");
//	out_str += 	"date,time,sighter,score" +
	out_str += 	"date,time,shooter,sighter,score" +
	",x"+xy_unit+",y"+xy_unit+",v"+speed_unit+
	",v_avg,v_sd,string,frame,distance(m)\n";
//	",v_avg,v_sd,string,shooter,frame,distance(m)\n";
	
	for (var i in report_strings) {
		var id = report_strings[i].id;
		if (id === undefined) {
			// new version only sends the ids... (have a look at download_csv in server.js)
			id = report_strings[i];
		}
		var string = module.exports.storage_read_string_sync(id);
		
		if (!string) {
			// console.log("string not found",id);
			_log.debug("string not found",id);
			continue; // old string may exist in archive but file is not accessible
		}
		// if this happens then archive is out of date and should be fixed at some point (todo)
		
		_log.debug("string found",id,string);
		// console.log("string found",id,"shooter=",report_strings[i].shooter,"string.shooter_count=",string.shooter_count);
		var frame = data.frames[string.frame_id];
		var face = data.faces[string.face_id];

		var first_shot = true;
		
		// for computing avg/sd
		var shot_data = [];
		var string_txt = "";
		var txt;
		
		try { // any exception caught while processing a string, will cause that string to be not displayed
			for (var j in string.shots) {
				var shot = string.shots[j];
				//console.log("shot",j,shot&&shot.valid, shot && shot.value, "shooternum=",shot && shot.shoooter_num);
				//console.log("shot",j,shot);
				if (!shot) continue;
				// only real shots for this shooter
				if (!shot.valid) continue;
				if (report_strings[i].shooter !== undefined && shot.shooter_num != report_strings[i].shooter) continue;
	
				txt = "";

				// time
				var date = new Date(shot.ts);
				var date_txt = "," + date.toLocaleTimeString();
				txt += date_txt + ",";
				
				if (string.shooter_count == 1) txt += "ONLY,";
				else if (string.shooter_count == 2) txt += (shot.shooter_num == 0)?("RIGHT,"):("LEFT,");
				else if (string.shooter_count == 3) txt += (shot.shooter_num == 0)?("RIGHT,"):((shot.shooter_num == 1)?("MIDDLE,"):("LEFT,"));

				txt += (shot.sighter?"S,":",") + shot.value + ",";
	
				var v_units = shot.v * speed_mpier;
				txt += (shot.x*xy_mpier).toFixed(xy_ndigs) + "," + (shot.y*xy_mpier).toFixed(xy_ndigs) + "," + v_units.toFixed(speed_ndigs);
				txt += "\n";
				// console.log("shot txt=",txt);
				string_txt += txt;
				
				if (!shot.sighter) shot_data.push({v: v_units});
			}
			// console.log("done string; string_txt=",string_txt);
			// console.log("done string; last shot==",shot);
			// done string; create header
			// toLocaleDateString() in nodejs is badly broken
			var date_txt = date.toDateString();
			txt = date_txt + ",,,,,,,,";
			if (shot_data.length >= 2) {
				// compute averages and sds
				var v_avg = 0;
				for (var k in shot_data) {
					v_avg += shot_data[k].v;
				}
				v_avg /= shot_data.length;
				var v_sd = 0;
				for (var k in shot_data) {
					v_sd += (shot_data[k].v - v_avg) * (shot_data[k].v - v_avg);
				}
				v_sd = Math.sqrt(v_sd / shot_data.length);
				txt += v_avg.toFixed(speed_ndigs) + "," + v_sd.toFixed(speed_ndigs+1) + ",";
			} else {
				txt += ",,"
			}
			txt += string.name.replace(",", ";") + ",";
			txt += frame.name + "," + face.distance.toFixed(0);
			txt += "\n";
			out_str += txt + string_txt + "\n"; // line break between strings
		} catch (e) {
			console.log("error string.shots",e);
			// ignore error; discard this string
		}
	}
	var out = fs.openSync(filename, "w");
	fs.writeSync(out, out_str);
	fs.closeSync(out);
}

module.exports.storage_write_shotslog = function (days, archive) {
	var start = _performance();

	var filename = "public/shotslog.csv";
	
	if (fs.existsSync("/tmp/")) { // there is /tmp, so use it instead (not the case for WINDOWS)
		var filename2 = "/tmp/shotslog.csv";
		rm_symlinkSync(filename);
		fs.symlinkSync(filename2, filename);
		filename = filename2;
	}
	
	var all_txt = "string,frame,face,mic_w,mic_h,x_cal,y_cal,seq,source,ts,date,shooter,valid,x,y,v,temp,m1,m2,m3,m4,m5,m6,m7,m8,micmap,measx,measy\n";
	var all_txt = "string,frame,face,mic_w(mm),mic_h(mm),x_cal(mm),y_cal(mm),source,ts,date(UTC),shooter,sighter,x(mm),y(mm),value,v(m/s),temp(C)\n";
	var all_txt = "string,frame,face,mic_w(mm),mic_h(mm),x_cal(mm),y_cal(mm),source,ts,date(UTC),shooter,sighter,x(mm),y(mm),value,v(m/s),temp(C)\n";
	var all_txt = "string,frame,face,mic_w(mm),mic_h(mm),x_cal(mm),y_cal(mm),source,ts,date(UTC),shooter#,realname,sighter,value,x(mm),y(mm),v(m/s),temp(C)\n";


	var string_id_arr = [];
	for (var i in archive) {
		string_id_arr.push(i);
	}
	string_id_arr.sort(function(a,b) {
	    if (archive[a].ts > archive[b].ts) return 1;
	    if (archive[a].ts < archive[b].ts) return -1;
	    return 0;
	});

	var ts_limit = 0;
	if (days && string_id_arr.length) { // only calc if there are strings
		ts_limit = archive[ string_id_arr[string_id_arr.length-1] ].ts - days * 1000 * 60 * 60 * 24;
	}


	for (var i in string_id_arr) {
		
		if (archive[string_id_arr[i]].ts < ts_limit) continue;
		
		var string = module.exports.storage_read_string_sync(string_id_arr[i]);
		
		if (!string) continue; // old string may exist in archive but file is not accessible
		// if this happens then archive is out of date and should be fixed at some point (todo)

		var frame = data.frames[string.frame_id] || {name: '--frame-deleted--'};
		var face = data.faces[string.face_id] || {name: '--face-deleted--'} ;
		
		var geom = frame.geom;
		if (string.geom) geom = string.geom; // use string record itself if exists
		
		var first_shot = true;

		for (var j in string.shots) {
			var shot = string.shots[j];
			if (!shot) continue;
			
			// for compatibility with older strings
			if (shot.shot) {
				shot.mics = shot.shot.counters;
				shot.num = shot.shot.shotnum;
			}

			var txt = "";

			// string/frame/face info on first shot only
			if (first_shot) {
				var strname = string.name;
				if ('string' != typeof strname) {
					strname = "";
				}
				strname = strname.replace(",", ";");
				txt = strname + "," + frame.name + "," + face.name + "," + trim_num(geom.mic_width,1) + "," +
					trim_num(geom.mic_height,1) + "," + trim_num(geom.x_cal,1) + "," + trim_num(geom.y_cal,1);
				all_txt += txt + "\n";
				first_shot = false;
    		}

   			txt = ",,,,,,,";

//			// seq, source, ts
//			txt += shot.num + "," + shot.source + "," + shot.ts + ",";
			//  source, ts
			txt += shot.source + "," + shot.ts + ",";

			// date
			var date = new Date(shot.ts);
			var date_txt = date.toDateString() + " " + date.toLocaleTimeString();
			txt += date_txt + ",";

			if (!shot.valid) {
				txt += ",";
			} else if (shot.shooter_num >= 0) {
				var shooter_num = shot.shooter_num;
				var shooter_name = "";
				if (string.use_role_shooter && string.role_shooter) {
					var uu = data.users[string.role_shooter[shooter_num]];
					if (uu) {
						shooter_num = "*"+uu.name;
						shooter_name = uu.realname || "";
					}
				}
				txt += shooter_num + "," + shooter_name + ",";
			} else {
				txt += "0,,";
			}

			txt += (shot.sighter == true) ? "sighter," : ",";
			if (shot.valid) {
				txt += shot.value + "," + trim_num(shot.x,1) + "," + trim_num(shot.y,1) + "," + trim_num(shot.v,1) + "," + trim_num(shot.tempC,1) + ",";
			} else {
				txt += ",,,,";
			}

			if (shot.valid) {
				all_txt += txt + "\n";
			} else {
				; // do not display not-valid shots
			}

		} // string.shots
		
	} // string_id_arr
	
	fs.writeFile(filename, all_txt, {flag: "w"}, function(err) {
		// no error handling
		console.log('storage_write_shotslog() done (' + _performance(start) + 'ms)', all_txt.length, 'bytes',"err=",err);
		storage_update_diskspace();
	});
}


// find and load an inactive string by id (for shots)
module.exports.storage_read_string_sync = function ( stringId ) {

	var fname;
	try {

		// read as zip
		var buf = fs.readFileSync( fname = "data/string-" + stringId + ".zip" );
		return JSON.parse( zlibjs.inflateSync( buf ) );

	} catch ( e ) {

		_log.error( "error reading string", stringId, "fname=", fname, e );
		// if zip fails, try as text (for compatibility with older strings)
		try {

			// read as human readable text
			var text = fs.readFileSync( fname = "data/string-" + stringId + ".txt" );
			return JSON.parse( text );

		} catch ( e ) {
			_log.error( "error reading string", stringId, "fname=", fname, e );
//console.log("DMC DEBUG - _log.error","error reading string", stringId, e );
		}

	}

}

// find and load an inactive string by id (for shots)
module.exports.storage_read_string_async = function ( string_id, callback_fp) {
	var start = _performance();
	var callback = callback_fp;
	if (typeof callback != 'function') callback = function(){}; // safety
	var filename_zip = "string-" + string_id + ".zip";
	var filename_txt = "string-" + string_id + ".txt";
	storage_read_string_fname_async(filename_zip,function (err, parsed_string) { // try .zip
		if (err) {
			if (err.code == 'ENOENT') { // .zip file not found; try ".txt"
				storage_read_string_fname_async(filename_txt,callback)
			} else { // report all other errors
				callback(err)
			}
		} else {
			callback(err,parsed_string);
		}
	});
}

// write an archived string to file - newer async version
module.exports.write_string = write_string = function (str) {
	var start = _performance();
	// console.log('new_storage.write_string - start (' + _performance(start) + 'ms)');
	try {
		// write zlibjs compressed
		var text = JSON.stringify(str, null, "\t");
		// console.log('new_storage.write_string - stringify done (' + _performance(start) + 'ms)',text.length,'bytes');
		var buffer = new Buffer(text);
		zlibjs.deflate(buffer, function (err, result) {
			// console.log('new_storage.write_string - deflate async callback start (' + _performance(start) + 'ms)');
			var check_text, zip;
			if (!err) {
				zip = result;
// TODO: make this async:
				check_text = zlibjs.inflateSync(zip).toString();
				// console.log('new_storage.write_string - inflateSync done (' + _performance(start) + 'ms)');
			}
			if (err || text != check_text) { // error zipping: write text
				// console.log("error: zipped string is not equivalent!");
				fs.writeFile("data/string-" + str._id + ".txt", text, function () {
					console.log('new_storage.write_string - writeFile async txt done (' + _performance(start) + 'ms)');
				});
			} else {
				fs.writeFile("data/string-" + str._id + ".zip", zip, function () {
					console.log('new_storage.write_string - writeFile async zip done (' + _performance(start) + 'ms)',zip.length,'bytes');
				});
			}
			// storage_update_diskspace();
			// console.log('new_storage.write_string - storage_update_diskspace done (' + _performance(start) + 'ms)');
		});
	} catch (e) {
		console.log("error writing string", str._id, e);
	}
}

function parse_string ( string_text, callback_fp) {
	var callback = callback_fp;
	var start = _performance();
	if (typeof callback != 'function') callback = function(){}; // safety
	try {
		var parsed_string = JSON.parse(string_text);
//		console.log('new_storage parse_string - JSON.parse done (' + _performance(start) + 'ms)', string_text.length, 'bytes');
		callback(null, parsed_string);
	} catch (e) {
		callback(e);
	}
}

// find and load an inactive string by filename
var storage_read_string_fname_async = function ( string_fname, callback_fp) {
	var start = _performance();
	var callback = callback_fp;
	if (typeof callback != 'function') callback = function(){}; // safety
	// read as zip
	fs.readFile("data/" + string_fname, function (err, read_buf) {
		console.log("storage_read_string_fname_async readFile",string_fname,"err=",err,"time=", _performance(start));
		if (err) {
			callback(err)
		} else {
//			console.log('storage_read_string - readFile done (' + _performance(start) + 'ms)', read_buf.length, 'bytes');
			// .zip needs uncompressing, otherwise .txt is fine as-is
			if (string_fname.match(/\.zip$/)) {
//_log.debug('zipfile',string_fname);
				zlibjs.inflate(read_buf, function (err, inflate_buf) {
					if (err) {
						callback(err);
					} else {
						parse_string(inflate_buf, callback);
					}
				});
			} else {
_log.debug('NOT zipfile',string_fname);
				parse_string(read_buf, callback);
			}
		}
	});
}


// get a unique ID for any new objects
var idCounter = 0; // todo: move this to file
function storage_get_unique_id( ) {
	return Date.now() + idCounter++;
}
module.exports.storage_get_unique_id = storage_get_unique_id;

// will import multiple target faces and images from a txt, tgt, or tgz file
module.exports.new_storage_import_targetfaces = function ( file, fclass_dualmode_supported, callback ) {
	
	var errors = [];
	// console.log("new_storage_import_targetfaces file=",file);

	// read a text file and import faces from it
	function parse_face_js( filepath, originalFilename ) {
		try {
			imported_targetfaces = require(filepath);
			if (! fclass_dualmode_supported) { // SOLO; create a "-F" target for F-Class zones
				for ( var i in imported_targetfaces ) {
					if ( imported_targetfaces[ i ].scoringzones_f !== undefined ) {
						console.log( "TARGETFACE '" + i + "' CONTAINS F_CLASS DEFINITIONS!" );
						var newF = JSON.parse( JSON.stringify( imported_targetfaces[ i ] ) );
						newF.scoringzones = newF.scoringzones_f;
						newF.scoringzones_f = undefined;
						imported_targetfaces[ i + "-F" ] = newF;
					}
					imported_targetfaces[ i ].scoringzones_f = undefined;
				}
			}
			face_import( imported_targetfaces );
			return true;
		} catch ( e ) {
			console.log( "Error reading file", e );
			errors.push( (originalFilename ? originalFilename : "" ) + e.toString() );
			return false;
		}
	}

	function is_textfile(filename) {
		return (filename.indexOf(".js") >= 0 ||
			filename.indexOf(".txt") >= 0 || filename.indexOf(".tgt") >= 0
		);
	}
	function is_tgzfile(filename) {
		return filename.indexOf(".tgz") >= 0
		|| filename.indexOf(".tar.gz") >= 0
		|| filename.indexOf(".gz") >= 0; // a bit desperate, but this seems to be the latest Microsoft EDGE "cuteness"
	}
	function is_tarfile(filename) { return filename.indexOf(".tar") >= 0; }
	function is_graphicsfile(filename) {
		return filename.indexOf(".png") >= 0
		|| filename.indexOf(".jpg") >= 0
		|| filename.indexOf(".jpeg") >= 0
		|| filename.indexOf(".gif") >= 0
		|| filename.indexOf(".svg") >= 0
		;
	}

	var imported_data = false;
	// imported file is a single text file with no images
	if (is_textfile(file.originalFilename)) {
		imported_data = parse_face_js( file.path, file.originalFilename ) || imported_data; // parse the one file
		fs.unlinkSync(file.path); // remove downloaded file
		if (imported_data) storage_write_faces( data ); // update to file
		broadcast_data(); // force full update to clients
		if ( 'function' == typeof callback ) {
			callback( ( errors.length === 0 ), errors );
		}
		
	// imported file is .tar with multiple text and images
	} else if (is_tgzfile(file.originalFilename) || is_tarfile(file.originalFilename)) {

		var cleanup_exec_cmd = "rm " + file.path; // to be run when we're finished, success or no

		// extract tgz file to temporary folder
		var tar_cmd = "rm -rf /tmp/data-temp &&  mkdir /tmp/data-temp";
		tar_cmd += "&& " + (is_tgzfile(file.originalFilename) ? "tar xzf" : "tar xf");
		tar_cmd += " " + file.path+ " -C /tmp/data-temp";
		console.log("tar_cmd",tar_cmd);

		exec(tar_cmd, null, function(error, stdout, stderr) {
			if (error) {
				errors.push("couldn't extract " + file.originalFilename);
			} else {
				
				var graphics_files_to_move = [];
				var dir = fs.readdirSync("/tmp/data-temp");
				for (var i in dir) {
					var subfile = dir[i];
					
					// parse each text file
					if (is_textfile(subfile)) {
						console.log("parsing faces: " + subfile);
						imported_data = parse_face_js( "/tmp/data-temp/" + subfile, subfile ) || imported_data;
					// move images into data folder
					// the client will access the image from a symlink in public
					} else if (is_graphicsfile(subfile)) {
						graphics_files_to_move.push(subfile)
					}
				}
				// move graphics files into place and then cleanup.
				var i, fq_str_graphics_files_to_move = "";
				for (i = 0; i < graphics_files_to_move.length; i++) {
					fq_str_graphics_files_to_move += " /tmp/data-temp/" + graphics_files_to_move[i];
				}
				if (fq_str_graphics_files_to_move.length) {
					cleanup_exec_cmd += "; mv " + fq_str_graphics_files_to_move + " data";
				}
			}
			cleanup_exec_cmd += "; rm -rf /tmp/data-temp";

			// regardless of success, cleanup, and then callback
			exec(cleanup_exec_cmd, function(error, stdout, stderr) {
				if (error) {
					errors.push("cleanup_exec_cmd error " + error);
				}
				if (imported_data) storage_write_faces( data ); // update to file
				if ( 'function' == typeof callback ) {
					callback( ( errors.length === 0 ), errors );
				}
			});
		}); // exec(tar_cmd ...
	} else { // file type not recognized
		errors.push(file.originalFilename + " file type not recognized");
		if ( 'function' == typeof callback ) {
			callback( ( errors.length === 0 ), errors );
		}
	}
}

// if a "name" member is missing, create it from "score"
function tweak1_scorezones(sz) {
	if (!Array.isArray(sz)) return;
	for (var i in sz) {
		var zone = sz[i];
		if ('undefined' == typeof zone.name && 'number' == typeof zone.score) {
			zone.name = '' + zone.score;
		}
	}
}


function face_import(face_data) {
	
	// load data into new target faces
	for (var i in face_data) {
		
		var id = null;
		
		// if exists, replace
		for (var j in data.faces) {
			if (data.faces[j].name == i) {
				id = j;
				break;
			}
		}
		
		// new id
		if (!id) {
			id = i; // 2020-07-20: use face name as id, instead ofstorage_get_unique_id()
			// DMC 29-Oct-2020, test first, too scared for now...
			id = storage_get_unique_id();
		}
		
		var face_to_import = face_data[i];

		//tweak1_scorezones(face_to_import.scoringzones);
		//tweak1_scorezones(face_to_import.scoringzones_f);
		
		// add to data
		data.faces[id] = face_to_import;
		data.faces[id].name = i;
		data.faces[id]._id = id;
		fixup_valfn(face_to_import);
	}
	
}


module.exports.backup_all_data = function backup_all_data(socket) {
	write_archive_compat(archive, function(err) {
		console.log("backup_all_data write_archive_compat err=",err);
		// regardless of failure of writing archive, create backup...
		exec("./create_backup.sh", function(error, stdout, stderr) {
			console.log("backup_all_data ./create_backup.sh error=",error);
			if (!error) {
				socket.emit("download_file", "backup.tgz");
				console.log("emit download file backup.tgz");
			} else {
				console.log("backup failed...", stderr);
			}
		});
	});
}




// write archive to file; now only used for backwards_compatibility
function write_archive_compat(archive_data, callback) {
	try {
		// var start = _performance();
		var str_archive = JSON.stringify(archive_data, null, "\t");
	    fs.writeFile("data/archive.txt", str_archive, function(error) { // for compat write archive.txt but not archive.bak
			if ('function' == typeof callback) {
				callback(error); // pass on error, if any
			}
        });
	} catch (e) {
		console.log("write_archive_compat Error",e,"writing to file");
		if ('function' == typeof callback) {
			callback(e);
		}
	}
}

// cloned from server_storage.js.... eventually need to delete it there
function write_file_and_backup (dirname, filename, data_string, callback) {
	var mainfilename = dirname+"/"+filename;
	var bak_filename = mainfilename + ".bak";
	var bak_delay = 5000; // write backup file after main file has had a chance to be flushed out
    fs.writeFile(mainfilename, data_string, function(error) {
//		console.log("setTimeout",bak_filename,bak_delay);
    	setTimeout(function() {
//		console.log("starting writing",bak_filename);
		    fs.writeFile(bak_filename, data_string, function() {
		console.log("finished writing",bak_filename);
		    }); // no callback
    	}, bak_delay);
    	
    	if (typeof callback == 'function') {
    		callback(error);
    	}
    	
    });

}


module.exports.storage_write_logfile = function (content, callback, filename/*optional*/) {
	if (!filename) {
		filename = (new Date()).toISOString() + '.html'; //'2020-10-29T17:19:48.568Z'
	}
	var filename_path = "data/" + filename;
	
	var data_symlink = "public/data";
	if (!fs.existsSync(data_symlink)) {
		fs.symlinkSync("../data", data_symlink);
	}

	fs.writeFile(filename_path, content, function(error) {
 		callback(error,filename_path);
 	});
}

function trim_num(nn,ndigs) { // type-safe .toFixed()
		if (typeof ndigs == 'undefined' || ndigs < 0) ndigs = 0;
		if (typeof nn == 'number') return +nn.toFixed(ndigs); // '+' needed to return 'number' not 'string'
		return nn;
}

